/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/cli/CliParser.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.cli;

import java.lang.annotation.Annotation;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.cli2.Argument;
import org.apache.commons.cli2.CommandLine;
import org.apache.commons.cli2.DisplaySetting;
import org.apache.commons.cli2.Group;
import org.apache.commons.cli2.Option;
import org.apache.commons.cli2.OptionException;
import org.apache.commons.cli2.builder.ArgumentBuilder;
import org.apache.commons.cli2.builder.CommandBuilder;
import org.apache.commons.cli2.builder.DefaultOptionBuilder;
import org.apache.commons.cli2.builder.GroupBuilder;
import org.apache.commons.cli2.commandline.Parser;
import org.apache.commons.cli2.option.Command;
import org.apache.commons.cli2.option.GroupImpl;
import org.apache.commons.cli2.validation.EnumValidator;

import com.netapp.common.MsgKey;
import com.netapp.common.cli.binding.BeanObjectBinding;
import com.netapp.common.cli.binding.BindingContainer;
import com.netapp.common.cli.binding.CommandBinding;
import com.netapp.common.cli.binding.FieldBinding;
import com.netapp.common.cli.binding.GlobalBindings;
import com.netapp.common.cli.binding.ObjectBinding;
import com.netapp.common.cli.binding.PrimitiveObjectBinding;
import com.netapp.common.cli.enumeration.CliEnumTranformerFactory;
import com.netapp.common.cli.enumeration.CliEnumTransformer;
import com.netapp.common.logging.MessageBundle;
import com.netapp.common.logging.NALogger;
import com.netapp.common.util.StringUtils;


/**
 * This class takes as input an annotated application programming interface (API) and produces a Command Line Interpreter (CLI)
 * that corresponds to that API. The generated CLI performs three main functions:
 * <ol>
 * <li>Parsing and error-checking the command line arguments
 * <li>Mapping the command and arguments to a method and parameters
 * <li>Invocation of the corresponding method in the API with the corresponding arguments
 * </ol>
 * The class relies heavily on reflection and annotation to implement this. 
 * @author sirianni
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public class CliParser {
    
    private static final NALogger logger = NALogger.getLogger(CliParser.class);

    private Parser parser;
    private Group commandGroup;
    private Map<com.netapp.common.cli.Command, Command> commandCodeMap;
    private Set<CommandBinding> commandBindings;
    private GlobalBindings globalBindings;
    private ArrayList<String> globalOptions;
    
    /**
     * If true then if a BEAN has only one nested argument do not require the switch for its setter.
     * Otherwise the switch for the setter is required.
     * SMO sets this to false. Defaults to true.
     */
    private boolean collapseSingleSetterBeans;
    
    /**
     * @deprecated The Command[] is no longer necessary.  Use the other constructor instead. 
     */
    @Deprecated
    public CliParser(com.netapp.common.cli.Command [] commands, Class<?> annotatedClass) {
        this(annotatedClass);
    }
    
    public CliParser(Class<?> annotatedClass) {
        this(annotatedClass, true);
    }
    
    /**
     * Construct a CliParser object from a class Annotated with the <code>@CliCommand</code> and <code>@CliParameter</code> annotations
     * @param collapseSingleSetterBeans - if true then if a BEAN has only one nested argument do not require the switch for its setter
     *      Otherwise the switch for the setter is required (SMO sets this to false) 
     */
    public CliParser(Class<?> annotatedClass, boolean collapseSingleSetterBeans) {
        this.collapseSingleSetterBeans = collapseSingleSetterBeans;
        this.commandCodeMap = new HashMap<com.netapp.common.cli.Command, Command>();
        this.commandBindings = new HashSet<CommandBinding>();
        CommandBuilder commandBuilder = new CommandBuilder();
        GroupBuilder commandGroupBuilder = new GroupBuilder().withMinimum(1).withMaximum(1);

        Map<String, com.netapp.common.cli.Command> commandNameMap = new HashMap<String, com.netapp.common.cli.Command>();
        
        /*
         * First group all the methods in the class by top-level CliCommand (e.g. "backup")
         */
        Map<com.netapp.common.cli.Command, Set<Method>> methodMap = new HashMap<com.netapp.common.cli.Command, Set<Method>>();
        for (Method method : annotatedClass.getMethods()) {
            CliCommand cliCommand = method.getAnnotation(CliCommand.class);
            if (cliCommand != null) {
                String commandString = cliCommand.command();
                com.netapp.common.cli.Command command = commandNameMap.get(commandString);
                if (command == null) {
                    command = new com.netapp.common.cli.Command(commandString);
                    commandNameMap.put(commandString, command);
                }
                String subCommand = cliCommand.subCommand();
                if (!subCommand.equals(com.netapp.common.cli.Command.NONE)) {
                    command.addSubCommand(new com.netapp.common.cli.Command(subCommand));
                }
                Set<Method> methodSet = methodMap.get(command);
                if (methodSet == null) {
                    methodSet = new HashSet<Method>();
                    if (null != command) {
                        methodMap.put(command, methodSet);
                    }
                    else {
                        logger.internal("Failed to find command for annotation '" + commandString + "'");
                    }
                }
                methodSet.add(method);
            }
        }
        
        /*
         * Now iterate over the top-level CliCommands and create an ApacheCLI Command object for each (e.g. "backup")
         */
        for (Map.Entry<com.netapp.common.cli.Command, Set<Method>> entry : methodMap.entrySet()) {
            Set<Method> methodSet = entry.getValue();
            com.netapp.common.cli.Command cmd = entry.getKey();
            
            Set<CommandBinding> bindingsForCommand = new HashSet<CommandBinding>();
            GroupBuilder subCommandGroupBuilder = new GroupBuilder();
            Group rootCommandChildren = null;
            
            /*
             * Iterate over all the methods for this top-level Command:
             *  - Create the corresponding ApacheCLI objects
             *  - Create Binding objects between the Methods and their corresponding ApacheCLI objects 
             */
            for (Method method : methodSet) {
                CommandBinding commandBinding = new CommandBinding();
                commandBinding.setMethod(method);
                this.commandBindings.add(commandBinding);
                bindingsForCommand.add(commandBinding);
                
                CliCommand cliCommand = method.getAnnotation(CliCommand.class);
                String commandName = cliCommand.command().toString();
                validateCommandLineToken(annotatedClass, method, commandName);
                
                /*
                 * Create a ApacheCLI Group object for each parameter in the method (e.g. "-profile")
                 */
                GroupBuilder methodCommandGroupBuilder = new GroupBuilder();
                int parameterNumber = 0;
                for (Annotation[] parameterAnnotations : method.getParameterAnnotations()) {
                    Class parameterClass = method.getParameterTypes()[parameterNumber++];
                    for (Annotation parameterAnnotation : parameterAnnotations) {
                        if (parameterAnnotation instanceof CliParameter) {
                            CliParameter cliParameter = (CliParameter) parameterAnnotation;
                            
                            String optionName = cliParameter.option();
                            validateCommandLineToken(annotatedClass, method, optionName);
                            
                            String argumentName = StringUtils.isEmpty(cliParameter.argumentName()) ? cliParameter
                                    .option() : cliParameter.argumentName();
                            Option option = this.optionFromParameter(
                                    commandBinding, cliParameter.option(),
                                    cliParameter.option(), argumentName, 
                                    "", cliParameter, parameterClass,
                                    parameterNumber, null);
                         
                            //                            Option option = this.optionFromParameter(commandBinding, cliParameter.option(), cliParameter.option(), cliParameter.option(), "", cliParameter, parameterClass, parameterNumber, null);
                            methodCommandGroupBuilder.withOption(option);
                        }
                    }
                }
                Group subCommandChildren = methodCommandGroupBuilder.create();
                
                /*
                 * If there is no SubCommand, then add the commandChildren directly to the top-level command
                 * Else, nest it within another ApacheCLI Command object and add that object to the top-level command
                 */
                if (cmd.getSubCommands().isEmpty() || cliCommand.subCommand().equals(com.netapp.common.cli.Command.NONE)) {
                    assert rootCommandChildren == null : "Can only have one method without a SubCommand per top-level Command";
                    rootCommandChildren = subCommandChildren;
                    subCommandGroupBuilder.withOption(rootCommandChildren);
                }
                else {
                    // validate that the given subCommand in the annotation matches a valid subCommand
                    String subCommandName = cliCommand.subCommand();
                    StringBuffer buffer = new StringBuffer();
                    boolean isValid = false;
                    
                    if (StringUtils.isEmpty(subCommandName)) {
                        isValid = true;
                    }
                    
                    for (com.netapp.common.cli.Command subCommand : cmd.getSubCommands()) {
                        if (subCommand.getName().equals(subCommandName)) {
                            isValid = true;
                            break;
                        }
                        
                        if (0 != buffer.length()) {
                            buffer.append(" ");
                        }
                        
                        buffer.append(subCommand.getName());
                    }
                    
                    if (!isValid) {
                        logger.error(MsgKey.CLI_BASE_REQUIRES_SUBCOMMAND, cmd.getName(), buffer.toString());
                    }
                    
                    Command subCommand = commandBuilder.withName(cliCommand.subCommand()).withChildren(subCommandChildren).create();
                    subCommandGroupBuilder.withOption(subCommand);
                    commandBinding.setSubCommand(subCommand);
                }
            }

            /*
             * Create a ApacheCLI Group object for all the SubCommands we just created, and finally create the 
             * top-level Command object with that subCommandGroup
             */
            Group commandChildren = subCommandGroupBuilder.create();
            /*
             * This description is important as it somehow allows Apache CLI to distinguish commands with the same top-level command:
             * e.g. "help" vs. "help backup"  
             */
            Command command = commandBuilder.withName(cmd.getName()).withDescription("Command").withChildren(commandChildren).create();
            
            for (CommandBinding commandBinding : bindingsForCommand) {
                commandBinding.setBaseCommand(command);
            }

            commandGroupBuilder.withOption(command);
            this.commandCodeMap.put(cmd, command);
        }
        
        this.commandGroup = commandGroupBuilder.create();
        
        /*
         * Now add "global" options (e.g. "-verbose", "-quiet")
         * Global options are defined as set methods on the annotatedClass that are annotated with a CliParameter annotation directly
         * on the method itself (the other CliParameter annotations are applied to method parameters)
         */
        GroupBuilder groupBuilder = new GroupBuilder();
        this.globalBindings = new GlobalBindings();
        for (CommandBinding commandBinding : this.commandBindings) {
            commandBinding.setGlobalBindings(this.globalBindings);
        }
        for (Method method : annotatedClass.getMethods()) {
            CliParameter cliParameter = method.getAnnotation(CliParameter.class);
            if (cliParameter != null) {
                assert method.getName().startsWith("set") : "Should only apply @CliParameter annotation to a set method";
                Class parameterClass = method.getParameterTypes()[0];
                String optionName = cliParameter.option();
                Option option = this.optionFromParameter(this.globalBindings, optionName, optionName, optionName, "", cliParameter, parameterClass, -1, method);
                groupBuilder.withOption(option);
            }
        }
        
        groupBuilder.withOption(this.commandGroup);
                
        /*
         * Create the ApacheCLI Parser object from the group of top-level commands 
         */
        this.parser = new Parser();
        this.parser.setGroup(groupBuilder.create());
        
        /*
         * Initialize the list of global options
         */
        initGlobalOptions();
    }

    public CommandBinding parse(String[] args) throws OptionException, BindingException {
        CommandLine commandLine = this.parseIntoCommandLine(args);
        Set<String> subCommands = new HashSet<String>();
        for (CommandBinding commandBinding : this.commandBindings) {
            Command command = commandBinding.getBaseCommand();
            Command subCommand = commandBinding.getSubCommand();
            if (subCommand != null) {
                if (commandLine.hasOption(command)) {
                    if (commandLine.hasOption(subCommand)) {
                        commandBinding.setCommandLine(commandLine);
                        return commandBinding;    
                    }
                    else {
                        subCommands.add(subCommand.getPreferredName());
                    }
                }
            }
        }
        for (CommandBinding commandBinding : this.commandBindings) {
            Command command = commandBinding.getBaseCommand();
            Command subCommand = commandBinding.getSubCommand();
            if (subCommand == null) {
                if (commandLine.hasOption(command)) {
                    commandBinding.setCommandLine(commandLine);
                    return commandBinding;
                }
            }
        }
        
        throw new BindingException(MessageBundle.localizeMessage(
            MsgKey.CLI_BASE_REQUIRES_SUBCOMMAND, args[0], subCommands.toString()));
    }
    
    /*
     * public for unit testing purposes.
     */
    public CommandLine parseIntoCommandLine(String[] args) throws OptionException {
        
        return this.parser.parse(moveGlobalOptionsToEnd(args));
    }

    /**
     * Initialize the set of global options for this parser
     */
    private void initGlobalOptions() {
        StringBuffer globalOption = new StringBuffer(16);
        
        List<FieldBinding> fieldBindings = this.globalBindings.getFieldBindings();

        if (globalOptions == null) {
            globalOptions = new ArrayList<String>(fieldBindings.size());
        }
        
        for (FieldBinding fieldBinding : fieldBindings) {
            globalOption.append('-');
            globalOption.append(fieldBinding.getObjectBinding().getCliParameter().option());
            
            globalOptions.add(globalOption.toString());
            
            globalOption.delete(0, globalOption.length());
        }        
    }
    
    /**
     * Rearrange the command line so global options are at the end
     * @param args the array of arguments
     * @return the reorganized arguments
     */
    private String[] moveGlobalOptionsToEnd(String[] args) {
        ArrayList<String> options = new ArrayList<String>(args.length);
        
        if (args.length > 0) {
            ArrayList<String> gOptions = new ArrayList<String>(globalOptions.size());
            boolean consumeThroughNextNonSwitch = false;
            
            for (String arg : args) {
                if (consumeThroughNextNonSwitch || globalOptions.contains(arg)) {
                    gOptions.add(arg);
                    // retrieve the object binding for this option and determine if the setter is a non-boolean
                    for (FieldBinding binding : this.globalBindings.getFieldBindings()) {
                        // binding is the current argument (plus '-' as the CliParameter option does not include it
                        if (arg.equals("-" + binding.getObjectBinding().getCliParameter().option())) {
                            // if there are more than two arguments to the method, it is complex and not just a boolean
                            if (1 < binding.getFieldSetter().getParameterTypes().length) {
                                consumeThroughNextNonSwitch = true;
                            }
                            else {
                                // should only be one in the list. If it is not a boolean, then consume the next non-switch argument
                                // as a global
                                for (Class<?> classType : binding.getFieldSetter().getParameterTypes()) {
                                    if (Boolean.class != classType && Boolean.TYPE != classType) {
                                        consumeThroughNextNonSwitch = true;
                                    }
                                }
                            }
                        }
                    }
                }
                else {
                    options.add(arg);
                }

                if (consumeThroughNextNonSwitch) {
                    if (!arg.startsWith("-")) {
                        consumeThroughNextNonSwitch = false;
                    }
                }
            }
            
            if (gOptions.size() > 0) {
                options.addAll(gOptions);
            }
        }
        
        String[] returnOptions = new String[options.size()];

        for (int i = 0; i < options.size(); i++) {
            returnOptions[i] = options.get(i);
        }
        
        return returnOptions; 
    }
    
    /**
     * Create an ApacheCLI Option object (and appropriate bindings) from an annotated method parameter
     * @param bindingContainer The binding object that is the parent of this parameter (Global/Command/Bean)
     */
    private Option optionFromParameter(BindingContainer bindingContainer, String optionName, String optionDescription, String argumentName, String prefix, CliParameter cliParameter, Class parameterClass, int parameterNumber, Method fieldSetter) {
        DefaultOptionBuilder defaultOptionBuilder = new DefaultOptionBuilder();
        
        defaultOptionBuilder.withShortName(optionName).withDescription(optionDescription);
        Option option;
        
        /*
         * This code is a bit ugly and hard to follow but it works...
         * 
         * The instanceof operator is kind of ugly.
         * I guess this could have been done using polymorphism, (i.e. having an addBinding() method on the 
         * BindingContainer interface instead of it being just a tag) but I didn't have time to refactor it.
         */
        switch (cliParameter.type()) {
            case BEAN:
            {
                BeanObjectBinding beanObjectBinding = new BeanObjectBinding(parameterClass, cliParameter);
                if (bindingContainer instanceof CommandBinding) {
                    ((CommandBinding) bindingContainer).new ParameterBinding(parameterNumber, beanObjectBinding);    
                }
                else if (bindingContainer instanceof BeanObjectBinding) {
                    ((BeanObjectBinding) bindingContainer).getFieldBindings().add(new FieldBinding(fieldSetter, beanObjectBinding));
                    
                    if (((BeanObjectBinding) bindingContainer).isParentIsArray()) {
                        beanObjectBinding.setParentIsArray(true);
                    }
                }
                else {
                    assert bindingContainer instanceof GlobalBindings;
                    ((GlobalBindings) bindingContainer).getFieldBindings().add(new FieldBinding(fieldSetter, beanObjectBinding));
                }
                
                // for an array class, we need to pass the type of bean it is, not the reference to the array itself
                if (parameterClass.isArray()) {
                    parameterClass = parameterClass.getComponentType();
                    beanObjectBinding.setParentIsArray(true);
                }
                
                /*
                 * This is a short cut for when only one setter is available and it is a primitive data type.
                 */
                if (this.collapseSingleSetterBeans && isSingleParameterBean(parameterClass)) {
                    Class<?> childClass = null;
                    Method setterMethod = null;
                    
                    for (Method method : parameterClass.getMethods()) {
                        if (method.getName().startsWith("set") && 1 == method.getParameterTypes().length) {
                            Class<?> [] types = method.getParameterTypes();
                            CliParameter childCliParameter = method.getAnnotation(CliParameter.class);
                            
                            if (isPrimitive(types[0]) && null != childCliParameter) {
                                childClass = types[0];
                                setterMethod = method;
                            }
                        }
                    }

                    option = buildPrimitiveOption(childClass, beanObjectBinding, cliParameter.prefix() + argumentName, defaultOptionBuilder, cliParameter);
                    
                    // even though we are shortcutting the name, we need to set a field binder.
                    ObjectBinding objectBinding = new PrimitiveObjectBinding(childClass, cliParameter, option);
                    beanObjectBinding.getFieldBindings().add(new FieldBinding(setterMethod, objectBinding));
                }             
                else {
                    Group optionGroup = this.groupFromBean(beanObjectBinding, cliParameter.prefix() + prefix, parameterClass);
                    option = defaultOptionBuilder.withChildren(optionGroup).withRequired(false).create();
                }
                
                beanObjectBinding.setOption(option);
                break;
            }
            case PRIMITIVE:
            case ENUM:
            case ARRAY:
            {
                option = buildPrimitiveOption(parameterClass, bindingContainer, argumentName, defaultOptionBuilder, cliParameter);
                
                ObjectBinding objectBinding;

                if (!isArrayType(bindingContainer, parameterClass)) {
                    objectBinding = new PrimitiveObjectBinding(parameterClass, cliParameter, option);
                }
                else {
                    objectBinding = new PrimitiveObjectBinding(parameterClass, cliParameter, option, true);
                }
                
                if (bindingContainer instanceof CommandBinding) {
                    ((CommandBinding) bindingContainer).new ParameterBinding(parameterNumber, objectBinding);    
                }
                else if (bindingContainer instanceof BeanObjectBinding) {
                    ((BeanObjectBinding) bindingContainer).getFieldBindings().add(new FieldBinding(fieldSetter, objectBinding));
                }
                else {
                    assert bindingContainer instanceof GlobalBindings;
                    ((GlobalBindings) bindingContainer).getFieldBindings().add(new FieldBinding(fieldSetter, objectBinding));
                }
                
                break;
            }
            default:
                throw new IllegalStateException();
        }

        return option;
    }

    private Group groupFromBean(BindingContainer bindingContainer, String prefix, Class beanClass) {
        GroupBuilder groupBuilder = new GroupBuilder();
        
        for (Method method : beanClass.getMethods()) {
            String methodName = method.getName();
            if (methodName.startsWith("set")) {
                Class parameterClass = method.getParameterTypes()[0];
                CliParameter cliParameter = method.getAnnotation(CliParameter.class);
                
                /*
                 * If the CliParameter is null, then the setter was not annotated
                 * In this case, it is not exposed via the CLI
                 */
                if (cliParameter != null) {
                    String optionName = cliParameter.option();
                    validateCommandLineToken(beanClass, method, optionName);
                    
                    String optionDescription = prefix + optionName;
                    groupBuilder.withOption(this.optionFromParameter(bindingContainer, optionName, optionDescription, optionDescription, prefix + cliParameter.prefix(), cliParameter, parameterClass, -1, method));                    
                }
            }
        }
        
        return groupBuilder.create();
    }
    
    /**
     * Check if the container type expects an array or if the current class is an array.
     * 
     * @param bindingContainer
     * @param parameterClass
     * @return if the container type expects an array or if the current class is an array
     */
    protected boolean isArrayType(BindingContainer bindingContainer, Class parameterClass) {
        if ((bindingContainer instanceof ObjectBinding
                && (((ObjectBinding) bindingContainer).isObjectAnArray() || ((ObjectBinding) bindingContainer).isParentIsArray()))
                || parameterClass.isArray()) {
            return true;
        }

        return false;
    }
    
    /**
     * Check if the class is an enum, or if the class is an array, if the component type is an enum.
     * 
     * @param parameterClass
     * @return if the class is an enum, or if the class is an array, if the component type is an enum
     */
    protected boolean isEnumType(Class parameterClass) {
        if (parameterClass.isEnum()) {
            return true;
        }
        
        if (parameterClass.isArray() && null != parameterClass.getComponentType() && parameterClass.getComponentType().isEnum()) {
            return true;
        }
        
        return false;
    }
    
    /**
     * Count how many methods the class has which start with 'set', contain only one argument
     * and the argument is a primitive type. The method also needs to have the CliParameter
     * set for the argument.
     * 
     * @param clazz
     * @return True if the bean class only has a single primitive parameter, false if it has
     *         two or more primitive members or one or more bean members
     */
    protected boolean isSingleParameterBean(Class<?> clazz) {
        int primitiveCount = 0;
        int beanCount = 0;
        
        for (Method method : clazz.getMethods()) {
            if (method.getName().startsWith("set") && 1 == method.getParameterTypes().length) {
                Class<?> [] types = method.getParameterTypes();
                CliParameter cliParameter = method.getAnnotation(CliParameter.class);
                
                if (null != cliParameter) {
                    if (isPrimitive(types[0])) {
                        primitiveCount++;
                    }
                    else {
                        beanCount++;
                    }
                }
            }
        }
        return (primitiveCount == 1 && beanCount == 0);
    }
    
    /**
     * To be a primitive for our useage, it must either be an actual
     * primitive class, an Enum, a String or a object that represents
     * a primitive.
     * 
     * @param clazz
     * @return True if the class is a primitive or primitive wrapper class
     */
    protected boolean isPrimitive(Class<?> clazz) {
        if (clazz.isPrimitive() || clazz.isEnum()) {
            return true;
        }
        
        if (clazz.isAssignableFrom(String.class) || clazz.isAssignableFrom(Integer.class)
                || clazz.isAssignableFrom(Long.class) || clazz.isAssignableFrom(Short.class) 
                || clazz.isAssignableFrom(Float.class) || clazz.isAssignableFrom(Double.class) 
                || clazz.isAssignableFrom(Boolean.class)) {
            return true;
        }
        
        return false;
    }
    
    protected Option buildPrimitiveOption(Class<?> parameterClass, BindingContainer bindingContainer, String argumentName, DefaultOptionBuilder defaultOptionBuilder, CliParameter cliParameter) {
        ArgumentBuilder argumentBuilder = new ArgumentBuilder();
        Option option = null;
        
        /*
         * If it's a boolean it doesn't have an argument.  
         * All other primitive types have an argument with Max and Min == 1
         */
        if (parameterClass.equals(Boolean.class) || parameterClass.equals(boolean.class)) {
            option = defaultOptionBuilder.withRequired(false).create();
        }
        else {
            argumentBuilder.withName(argumentName).withMinimum(1);
            
            // non-arrays only have one value
            if (!isArrayType(bindingContainer, parameterClass)) {
                argumentBuilder.withMaximum(1);
            } 
            
            // enumerations needs to have all possible values set as a validator
            if (isEnumType(parameterClass)) {
                
                Set<String> enumSet = new HashSet<String>();
                Object[] enumValues = null;
                
                // handle arrays of enums. We only want to check if the enum is an array, not the binding container
                if (parameterClass.isArray()) { 
                    enumValues = parameterClass.getComponentType().getEnumConstants();
                } 
                else {
                    enumValues = parameterClass.getEnumConstants(); 
                }
                
                CliEnumTransformer cliEnumTransformer = CliEnumTranformerFactory.createCliEnumTransformer(cliParameter.enumTransformer());

                for (Object value : enumValues) {
                    enumSet.add(cliEnumTransformer.enumToCliOption((Enum) value));
                }
                
                EnumValidator validator = new EnumValidator(enumSet);
                
                argumentBuilder.withValidator(validator);
            }
            
            Argument argument = argumentBuilder.create();
            option = defaultOptionBuilder.withArgument(argument).withRequired(false).create();
        }

        return option;
    }


    
    public Parser getParser() {
        return this.parser;
    }
    
    public Group getCommandGroup() {
        return this.commandGroup;
    }
    
    public Map<com.netapp.common.cli.Command, Command> getCommandCodeMap() {
        return this.commandCodeMap;
    }
    
    protected Set<CommandBinding> getCommandBindings() {
        return this.commandBindings;
    }
    
    /**
     * Get the usage for the entire parser
     * @param shellCommand The shell command to prefix to the usage (e.g. "smo")
     */
    public String getUsage(String shellCommand) {
        StringBuffer usage = new StringBuffer();
        usage.append(shellCommand + " ");
        Set<DisplaySetting> displaySettings = new HashSet<DisplaySetting>();
        this.commandGroup.appendUsage(usage, displaySettings, null);
        return usage.toString();
    }

    /**
     * Get the usage for a top-level command
     * @param shellCommand The shell command to prefix to the usage (e.g. "smo")
     * @param command The command for which usage is desired
     */
    @SuppressWarnings("unchecked") // Generics casts from Apache Commons
    public String getUsage(String shellCommand, com.netapp.common.cli.Command command) {
        
        Set<DisplaySetting> displaySettings = new HashSet<DisplaySetting>();
        displaySettings.add(DisplaySetting.DISPLAY_PARENT_CHILDREN);
        displaySettings.add(DisplaySetting.DISPLAY_ARGUMENT_BRACKETED);
        displaySettings.add(DisplaySetting.DISPLAY_GROUP_EXPANDED);
        displaySettings.add(DisplaySetting.DISPLAY_PARENT_ARGUMENT);

        Command cliCommand = this.commandCodeMap.get(command);

        StringBuffer usage = new StringBuffer();
        if (cliCommand != null) {
            GroupImpl groupImpl = (GroupImpl) cliCommand.getChildren();
            List<Option> options = groupImpl.getOptions();
            for (Option option : options) {
                usage.append(shellCommand + " " + cliCommand.getPreferredName() + " ");
                option.appendUsage(usage, displaySettings, null);
                usage.append('\n');
            }
        }
        else {
            usage.append("The command \"").append(command.getName().toLowerCase()).append("\" is not valid.");
        }

        return usage.toString().replace('|', ' ');
  
    }
    
    private void validateCommandLineToken(Class clazz, Method method, String tokenName) {
        if (tokenName.contains("_")) {
            String msg = logger.internal("Bad command line token name: %s Class: %s Method: %s", tokenName, clazz.getName(), method.getName());
            throw new RuntimeException(msg);
        }
    }

    public boolean isCommandValid(com.netapp.common.cli.Command command) {
        return this.commandCodeMap.get(command) != null;
    }

}
