/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/cli/binding/BeanObjectBinding.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.cli.binding;

import java.lang.reflect.Array;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.List;

import org.apache.commons.cli2.CommandLine;

import com.netapp.common.MsgKey;
import com.netapp.common.cli.CliParameter;
import com.netapp.common.cli.enumeration.CliEnumTranformerFactory;
import com.netapp.common.cli.enumeration.CliEnumTransformer;
import com.netapp.common.cli.validator.ValidationException;
import com.netapp.common.cli.validator.ValidatorType;
import com.netapp.common.logging.NALogger;

public class BeanObjectBinding extends ObjectBinding implements BindingContainer {
    private static final NALogger logger = NALogger.getLogger(BeanObjectBinding.class);
    private List<FieldBinding> fieldBindings;

    public BeanObjectBinding(Class<?> beanClass, CliParameter cliParameter) {
        super(beanClass, cliParameter);
        this.fieldBindings = new ArrayList<FieldBinding>();
    }

    public List<FieldBinding> getFieldBindings() {
        return this.fieldBindings;
    }
    
    @Override
    protected void validate(CommandLine commandLine) throws ValidationException {
        for (ValidatorType validatorType : this.getCliParameter().validators()) {
            validatorType.getValidator().validate(this, commandLine);
        }
    }
    
    @Override
    public Object doHydrate(CommandLine commandLine) throws ValidationException {
        try {
            if (this.isNull(commandLine)) {
                return null;
            }
            
            if (isObjectAnArray() || isParentIsArray()) {
                return doArrayHydrate(commandLine);
            }
            
            Object object = getObjectClass().newInstance();
            
            for (FieldBinding fieldBinding : this.fieldBindings) {
 
                Object value = fieldBinding.hydrate(commandLine);
                
                /**
                 * Skip logic in order to avoid NPE if optional  
                 * cli arguments are supplied. 
                 */
                if (value != null) {                
                    if (fieldBinding.getObjectBinding().getObjectClass().isEnum()) {
                        value = getEnumValue(fieldBinding, value);
                    }
                    try {
                        fieldBinding.fieldSetter.invoke(object, value);
                    }
                    catch (IllegalArgumentException e) {
                        logger.debug(MsgKey.CLI_BASE_DEBUG_BINDING_BEAN_PROPERTY, 
                            getObjectClass().getSimpleName(), fieldBinding.fieldSetter.getName(),
                            (value != null ? value.getClass().getSimpleName() : null), value,
                            e.getLocalizedMessage(), e);
                        throw new ValidationException(
                            logger.error(MsgKey.CLI_BASE_ERROR_BINDING_BEAN_PROPERTY, 
                                getObjectClass().getSimpleName(), fieldBinding.fieldSetter.getName(),
                                (value != null ? value.getClass().getSimpleName() : null), value,
                                e.getLocalizedMessage(), e));
                    }
                }
            }
            
            return object;
        }
        catch (InstantiationException e) {
            throw new RuntimeException(e);
        }
        catch (IllegalAccessException e) {
            throw new RuntimeException(e);
        }
        catch (InvocationTargetException e) {
            throw new RuntimeException(e.getCause());
        }
    }

    @SuppressWarnings("unchecked")
    public Object doArrayHydrate(CommandLine commandLine) throws ValidationException, InstantiationException,
        IllegalAccessException, InvocationTargetException {

        List<Object> objects = new ArrayList<Object>();
        boolean firstTimeThrough = true;
        
        for (FieldBinding fieldBinding : this.fieldBindings) {
            
            Object value = fieldBinding.hydrate(commandLine);
            
            /**
             * Skip logic in order to avoid NPE if optional bean type 
             * cli arguments are supplied. 
             */
            if (value != null) {
                List<Object> values = null;

                if (value.getClass().isArray() || value.getClass().isInstance(Array.class)) {
                    values = new ArrayList<Object>();

                    for (int i = 0; i < Array.getLength(value); i++) {
                        values.add(Array.get(value, i));
                    }
                }
                else {
                    values = (List<Object>) value;
                }


                for (int i = 0; i < values.size(); i++) {
                    Object currentValue = values.get(i);
                    Object object = null;

                    if (firstTimeThrough) {
                        Class classType = getObjectInstanceType();

                        object = classType.newInstance();
                        objects.add(object);

                    }
                    else {
                        object = objects.get(i);
                    }

                    /*
                System.out.println("  Bean: " + getCliParameter().argumentName());
                System.out.println("  Class: " + getObjectClass().getName());
                System.out.println("  Object Class: " + object.getClass().getName());
                System.out.println("  Value Class: " + currentValue.getClass().getName());
                System.out.println("  Value: " + currentValue);
                     */

                    if (fieldBinding.getObjectBinding().getObjectClass().isEnum()) {
                        currentValue = getEnumValue(fieldBinding, currentValue);
                    }

                    fieldBinding.fieldSetter.invoke(object, currentValue);
                }
                firstTimeThrough = false; 
            }
        }
        
        /*
         * Just returning objects.toArray() is not valid. We actually need the Array object type to be
         * created and each element set.
         */
        Object[] objArray = objects.toArray();
        Object arrayInstance = Array.newInstance(getObjectInstanceType(), objArray.length);
        
        for (int i = 0; i < objArray.length; i++) {
            Array.set(arrayInstance, i, objArray[i]);
        }
        
        return arrayInstance;
    }
    
    @SuppressWarnings("unchecked")
    protected Enum<?> getEnumValue(FieldBinding fieldBinding, Object value) throws InstantiationException, IllegalAccessException {  
        CliEnumTransformer cliEnumTransformer = CliEnumTranformerFactory.createCliEnumTransformer(fieldBinding.getObjectBinding().getCliParameter().enumTransformer());
        Class<?> clazz = fieldBinding.getObjectBinding().getObjectClass();
        return cliEnumTransformer.cliOptionToEnumValue(clazz.asSubclass(Enum.class), value.toString());
    }

    @Override
    public String toString() {
        return getObjectClass() + ":(" + this.fieldBindings.toString() + ")";
    }
    
    protected Class<?> getObjectInstanceType() {
        Class<?> classType = getObjectClass();
        
        if (null != classType.getComponentType()) {
            classType = classType.getComponentType();
        }
        
        return classType;
    }
}