/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/credentials/FileCredentialPersistenceManager.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc. 
 * All rights reserved.
 */
package com.netapp.common.credentials;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.List;

import javax.crypto.Cipher;
import javax.crypto.CipherInputStream;
import javax.crypto.CipherOutputStream;

import org.ho.yaml.Yaml;
import org.ho.yaml.YamlEncoder;

import com.netapp.common.MsgKey;
import com.netapp.common.encryption.EncryptionException;
import com.netapp.common.encryption.SymmetricEncryption;
import com.netapp.common.logging.NALogger;
import com.netapp.common.util.StringUtils;

/**
 * Persist the credential information to the file system. The file allows for future additions
 * to the file format in order to preserve backward compatibility.
 * 
 * The following configuration properties need to be set by the application
 * 
 * <ul>
 *   <li>credential.persistence.file.path
 * </ul>
 * 
 * For file persistence, this class uses YAML as the file format. The JYaml project provides
 * the YAML implementation to allow us to persist piece of credential information. By default,
 * this information would be available in plain text at the persistentFilePath location. In 
 * order to keep prying eyes away from this file, the file is encrypted using an internal, static
 * key.
 * 
 * @author joshb
 *
 */
public class FileCredentialPersistenceManager extends CredentialPersistenceManager {
    private static final NALogger LOGGER = NALogger.getLogger(FileCredentialPersistenceManager.class);

    private static final String PERSISTENCE_FILE_ALGORITHM = "AES";
    private static final byte[] PERSISTENCE_FILE_SECRET_KEY = { (byte) 26, (byte) -37, (byte) -85, (byte) 36,
        (byte) 98, (byte) -22, (byte) 85, (byte) -44, (byte) -77, (byte) -30, (byte) 56, (byte) -12, (byte) 70,
        (byte) -115, (byte) -50, (byte) -27 };

    private String persistentFilePath;
    private List<PersistentFileCredential> credentialList;

    public FileCredentialPersistenceManager() {
        credentialList = new ArrayList<PersistentFileCredential>();
    }
    
    /** 
     * Init method. To be called after all configuration is complete, generally
     * by Spring or another framework.
     *
     */
    public void init() throws CredentialPersistenceException {
        super.init();
    }

    /**
     * 
     * @see com.netapp.common.credentials.CredentialPersistenceManager#persistCredentialToStore(com.netapp.common.credentials.Credential, java.lang.String)
     */
    @Override
    protected synchronized void persistCredentialToStore(Credential credential, String type) throws CredentialPersistenceException {
        /*
         * If we already have an entry for the password, remove it.
         */
        if (null != retreivePasswordFromStore(credential.getKey(), type)) {
            LOGGER.debug(MsgKey.CREDENTIAL_PERSISTENCE_MANAGER_FILE_INFO_READDING_CREDENTIAL, credential.getKey()
                .getKey(), type);
            removeCredential(credential.getKey(), type);
        }

        PersistentFileCredential persistentCredential = new PersistentFileCredential();

        persistentCredential.setKey(credential.getKey());
        persistentCredential.setType(type);
        persistentCredential.setPassword(encrypt(credential.getPassword()));
        
        loadCredentialStorage();

        addCredentialToList(persistentCredential);

        writeCredentialStorage();
    }

    /**
     * 
     * @see com.netapp.common.credentials.CredentialPersistenceManager#removeCredentialFromStore(com.netapp.common.credentials.CredentialKey, java.lang.String)
     */
    @Override
    protected synchronized void removeCredentialFromStore(CredentialKey key, String type) throws CredentialPersistenceException {
        PersistentFileCredential persistentCredential = new PersistentFileCredential();

        persistentCredential.setKey(key);
        persistentCredential.setType(type);

        loadCredentialStorage();

        if (containsCredential(persistentCredential)) {
            removeCredentialFromList(persistentCredential);
            writeCredentialStorage();
        }
        else {
            LOGGER.info(MsgKey.CREDENTIAL_PERSISTENCE_MANAGER_FILE_INFO_CREDENTIAL_DOES_NOT_EXIST, key.getKey(), type);
        }
    }

    /**
     * 
     * @see com.netapp.common.credentials.CredentialPersistenceManager#retreivePasswordFromStore(com.netapp.common.credentials.CredentialKey, java.lang.String)
     */
    @Override
    protected synchronized String retreivePasswordFromStore(CredentialKey key, String type) throws CredentialPersistenceException {
        String password = null;
        PersistentFileCredential persistentCredential = new PersistentFileCredential();

        persistentCredential.setKey(key);
        persistentCredential.setType(type);

        password = getPasswordFromCredentialList(persistentCredential);

        if (null == password) {
            loadCredentialStorage();

            password = getPasswordFromCredentialList(persistentCredential);
        }

        return password;
    }
    
    /**
     * 
     * @see com.netapp.common.credentials.CredentialPersistenceManager#listCredentialsFromStore(java.lang.String)
     */
    @Override
    protected synchronized List<Credential> listCredentialsFromStore(String type) throws CredentialPersistenceException {
        // force all credentials to be loaded
        loadCredentialStorage();
        
        List<Credential> credentials = new ArrayList<Credential>();
        
        for (PersistentFileCredential current : getCredentialList()) {
            /*
             * If type is null, add all entries. Otherwise, if type is equal
             * to the type stored, add the entry.
             */
            if (StringUtils.isEmpty(type) || current.getType().equals(type)) {
                Credential credential = new Credential(current.getKey(), "");
                
                credentials.add(credential);
            }
        }
        
        return credentials;
    }

    /**
     * 
     * @see com.netapp.common.credentials.CredentialPersistenceManager#loadConfiguration()
     */
    @Override
    protected void loadConfiguration() throws CredentialPersistenceException {
        super.loadConfiguration();

        if (StringUtils.isEmpty(getPersistentFilePath())) {
            String filePath = CredentialConstants.PERSISTENCE_FILE_PATH.lookup(getContext());

            if (null == filePath || StringUtils.isEmpty(filePath)) {
                throw new CredentialPersistenceException(
                    LOGGER
                        .error(MsgKey.CREDENTIAL_PERSISTENCE_MANAGER_FILE_CONFIGURATION_ERROR_PERSISTENT_FILE_PATH_NOT_DEFINED));
            }

            setPersistentFilePath(filePath);
        }
    }

    /**
     * Write all entries from the internal credential list to the output file. This method
     * is synchronized so that we cannot have two process attempting to write out the same
     * file at the same time.
     * 
     * @throws CredentialPersistenceException
     */
    protected synchronized void writeCredentialStorage() throws CredentialPersistenceException {
        File file = new File(getPersistentFilePath());

        /*
         * If we have no data to actually write, then just remove the file and exit
         */
        if (getCredentialList().isEmpty()) {
            if (file.exists()) {
                file.delete();
            }

            LOGGER.debug(MsgKey.CREDENTIAL_PERSISTENCE_MANAGER_FILE_INFO_EMPTY_CREDENTIAL_LIST);

            return;
        }

        YamlEncoder encoder = null;

        try {
            Cipher cipher = SymmetricEncryption.getCipher(PERSISTENCE_FILE_ALGORITHM, PERSISTENCE_FILE_SECRET_KEY,
                Cipher.ENCRYPT_MODE);
            OutputStream stream = new CipherOutputStream(new FileOutputStream(file), cipher);

            encoder = new YamlEncoder(stream);

            for (PersistentFileCredential current : getCredentialList()) {
                encoder.writeObject(current);
            }

            LOGGER.debug(MsgKey.CREDENTIAL_PERSISTENCE_MANAGER_FILE_INFO_PERSISTED_CREDENTIALS, getCredentialList()
                .size());

            getCredentialList().clear();
        }
        catch (FileNotFoundException e) {
            throw new CredentialPersistenceException(e);
        }
        catch (EncryptionException e) {
            throw new CredentialPersistenceException(e);
        }
        finally {
            if (null != encoder) {
                encoder.close();
            }
        }
    }

    /**
     * Read all entries from persistent file.
     * 
     * @throws CredentialPersistenceException
     */
    protected synchronized void loadCredentialStorage() throws CredentialPersistenceException {
        File file = new File(getPersistentFilePath());

        if (!file.exists() || !file.isFile() || 0 == file.length()) {
            LOGGER.warn(MsgKey.CREDENTIAL_PERSISTENCE_MANAGER_FILE_INFO_PERSISTENT_FILE_DOES_NOT_EXIST, getPersistentFilePath());
            return;
        }

        try {
            Cipher cipher = SymmetricEncryption.getCipher(PERSISTENCE_FILE_ALGORITHM, PERSISTENCE_FILE_SECRET_KEY,
                Cipher.DECRYPT_MODE);
            InputStream stream = new CipherInputStream(new FileInputStream(file), cipher);

            getCredentialList().clear();

            for (PersistentFileCredential current : Yaml.loadStreamOfType(stream, PersistentFileCredential.class)) {
                getCredentialList().add(current);
            }

            LOGGER.debug(MsgKey.CREDENTIAL_PERSISTENCE_MANAGER_FILE_INFO_RETRIEVED_CREDENTIALS, getCredentialList()
                .size());
        }
        catch (EncryptionException e) {
            throw new CredentialPersistenceException(e);
        }
        catch (FileNotFoundException e) {
            throw new CredentialPersistenceException(e);
        }
    }

    /**
     * If the credential is in the current list maintained in RAM, decrypt and return the password. Otherwise, return null.
     * 
     * @param persistentCredential
     * @return Unencrypted password for the supplied credential
     * @throws CredentialPersistenceException
     */
    protected synchronized String getPasswordFromCredentialList(PersistentFileCredential persistentCredential)
        throws CredentialPersistenceException {
        String password = null;
        // This should be synchronized on a more granular lock, but since
        // we can reload the underlying file at any time
        for (PersistentFileCredential current : getCredentialList()) {
            if (equalCredentials(current, persistentCredential))  {
                return decrypt(current.getPassword());
            }
        }
        return null;
    }
    
    /**
     * In order to support optional case insensitive matches, we can no longer rely on the PersistentFileCredential.equals
     * method. To get around it, this helper method will loop through the credentials and check each on individually while
     * performing the proper case check based on the configuration.
     * 
     * @param credential
     * @return <code>true</code> if the current credential store contains an entry for <code>credential</code>, <code>false</code> otherwise.
     */
    protected synchronized boolean containsCredential(PersistentFileCredential credential) {
        for (PersistentFileCredential current : getCredentialList()) {
            if (equalCredentials(current, credential)) {
                return true;
            }
        }
        
        return false;
    }
    
    /**
     * Perform the actual case (in)sensitivity match based on the current configuration.
     * 
     * @param credential1
     * @param credential2
     * @return <code>true</code> if <code>credential1</code> equals <code>credential2</code> based on the current case-sensitivity setting.
     */
    protected boolean equalCredentials(PersistentFileCredential credential1, PersistentFileCredential credential2) {
        if (isCaseInsensitive() && credential1.equalsIgnoreCase(credential2)) {
            return true;
        }
        else if (!isCaseInsensitive() && credential1.equals(credential2)) {
            return true;
        }
        
        return false;
    }
    
    /**
     * In order to support optional case insensitive matches, we can no longer rely on the PersistentFileCredential.equals
     * method. To get around it, this helper method will loop through the credentials and check each on individually while
     * performing the proper case check based on the configuration in order to determine their position in the list.
     * 
     * If the credential cannot be found, -1 is returned
     * 
     * @param credential
     * @return index of credential in the credential list if found, otherwise -1
     */
    protected synchronized int indexOfCredential(PersistentFileCredential credential) {
        for (int i = 0; i < getCredentialList().size(); i++) {
            PersistentFileCredential current = getCredentialList().get(i);
            
            if (equalCredentials(current, credential)) {
                return i;
            }
        }
        
        return -1;
    }
    
    /**
     * A helper method to add a credential to our internal list. This method must handle to potential case
     * insensitivity of the key.
     * 
     * @param credential
     */
    protected synchronized void addCredentialToList(PersistentFileCredential credential) {
        int index = indexOfCredential(credential);
        if (index != -1) {
            getCredentialList().get(index).setPassword(credential.getPassword());
        }
        else {
            getCredentialList().add(credential);
        }
    }
    
    /**
     * A helper method to remove a credential from our internal list. This method must handle to potential case
     * insensitivity of the key.
     * 
     * @param credential
     */
    protected synchronized void removeCredentialFromList(PersistentFileCredential credential) {
        int index = indexOfCredential(credential);
        if (index != -1) {
            getCredentialList().remove(index);
        }
    }

    protected String getPersistentFilePath() {
        return persistentFilePath;
    }

    public void setPersistentFilePath(String persistentFilePath) {
        this.persistentFilePath = persistentFilePath;
    }

    protected List<PersistentFileCredential> getCredentialList() {
        return credentialList;
    }

    protected void setCredentialList(List<PersistentFileCredential> credentialList) {
        this.credentialList = credentialList;
    }
}
