/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/flow/ForLoopTemplate.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.flow;

import static com.netapp.common.flow.FlowConstants.*;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

import org.w3c.dom.Node;

import com.netapp.common.MsgKey;
import com.netapp.common.logging.NALogger;
import com.netapp.common.util.ClassUtils;
import com.netapp.common.xml.DOMUtils;

/**
 * Represents a for loop tag in the tasks xml.
 * A for loop iterates over the values in a collection and exposes each value to an enclosed task
 * or nested for loop in turn.
 * 
 * @author sunday
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
class ForLoopTemplate implements StackFrameTemplate, ExecutionStep {
    private static final NALogger logger = NALogger.getLogger(ForLoopTemplate.class);
    
    private StackFrameTemplate enclosingFrameTemplate;
    private List<ExecutionStep> executionSteps;
    private RegisterTemplate values;
    private Method[] propertyPath;
    private RegisterTemplate var;
    private TaskInstanceFactory taskInstanceFactory;
    
    ForLoopTemplate(StackFrameTemplate enclosingFrameTemplate, Node xml, TaskInstanceFactory taskInstanceFactory) {
        this.enclosingFrameTemplate = enclosingFrameTemplate;
        this.taskInstanceFactory = taskInstanceFactory;
        String valuesAttr = DOMUtils.getAttribute(xml, VALUES_ATTRIBUTE);
        String varAttr = DOMUtils.getAttribute(xml, VAR_ATTRIBUTE);
        
        String[] registerPath = valuesAttr.split("\\.");
        String valuesName = registerPath[0];
        
        this.values = enclosingFrameTemplate.getRegisterInstanceTemplate(valuesName);
        if (this.values == null) {
            throw new FlowException(logger.error(MsgKey.FLOW_FOR_VALUES_NOT_DEFINED, valuesName));
        }
        
        Type registerType = values.getType();
        
        if (registerPath.length > 1) {
            propertyPath = new Method[registerPath.length - 1];
            for (int i = 1; i < registerPath.length; i++) {
                Class<?> clazz = ClassUtils.extractImmediateClass(registerType);
                Method m = ClassUtils.findGetterMethod(clazz, registerPath[i]);
                if (m == null) {
                    throw new FlowException(logger.error(MsgKey.FLOW_INPUT_PROPERTY_NOT_FOUND, registerPath[i], valuesAttr));
                }
                registerType = ClassUtils.extractFieldTypeFromBeanMethod(m);
                propertyPath[i - 1] = m;
            }
        }
        
        Class<?> clazz = ClassUtils.extractCollectionClass(registerType);
        if (clazz == null || !(registerType instanceof ParameterizedType)) {
            throw new FlowException(logger.error(MsgKey.FLOW_FOR_VALUES_NOT_COLLECTION, valuesAttr));
        }
        
        registerType = ((ParameterizedType)registerType).getActualTypeArguments()[0];
        this.var = new RegisterTemplate(varAttr, registerType);
        
        this.executionSteps = new LinkedList<ExecutionStep>();
        for (Node child = xml.getFirstChild(); child != null; child = child.getNextSibling()) {
            if (Node.ELEMENT_NODE != child.getNodeType()) {
                continue;
            }
            if (DOMUtils.isElement(child, WORKFLOW_SCHEMA_NAMESPACE, FOR_ELEMENT)) {
                this.executionSteps.add(new ForLoopTemplate(this, child, this.taskInstanceFactory));
            }
            else {
                this.executionSteps.add(new TaskInstanceTemplate(this, child, this.taskInstanceFactory));    
            }
        }
    }

    /**
     * @see com.netapp.common.flow.StackFrameTemplate#getRegisterInstanceTemplate(java.lang.String)
     */
    public RegisterTemplate getRegisterInstanceTemplate(String name) {
        if (name.equals(this.var.getName())) {
            return this.var;
        }
        return this.enclosingFrameTemplate.getRegisterInstanceTemplate(name);
    }

    public void execute(StackFrame context) throws Throwable {
        Object obj = context.getRegisterValue(values.getName());
        if (obj == null) {
            logger.internal("For loop values is null");
            return;
        }
        if (propertyPath != null) {
            for (Method m : propertyPath) {
                Exception ex = null;
                try {
                    obj = m.invoke(obj);
                    if (obj == null) {
                        logger.internal("For loop value property from %s is null", m.getName());
                        return;
                    }
                }
                catch (IllegalArgumentException e) {
                    ex = e;
                }
                catch (IllegalAccessException e) {
                    ex = e;
                }
                catch (InvocationTargetException e) {
                    ex = e;
                }
                if (ex != null) {
                    throw new FlowException(logger.error(MsgKey.FLOW_INPUT_PROPERTY_REFLECTION_EXCEPTION, m.getName(), obj.getClass().getName()));
                }
            }
        }
        
        Collection<?> col = (Collection<?>)obj;
        ForLoopFrame frame = new ForLoopFrame(context);
        for (Object collectionItem : col) {
            frame.setValue(collectionItem);
            for (ExecutionStep executionStep : this.executionSteps) {
                executionStep.execute(frame);    
            }
        }
    }
    
    private class ForLoopFrame implements StackFrame {
        private StackFrame enclosingFrame;
        private Object value;
        
        protected ForLoopFrame(StackFrame enclosingFrame) {
            this.enclosingFrame = enclosingFrame;
        }
        
        protected void setValue(Object value) {
            this.value = value;
        }
        
        public OperationContext getContext() {
            return enclosingFrame.getContext();
        }

        public Object getRegisterValue(String name) {
            if (name.equals(var.getName())) {
                return value;
            }
            return this.enclosingFrame.getRegisterValue(name);
        }

        public void setRegisterValue(String name, Object obj) {
            this.enclosingFrame.setRegisterValue(name, obj);
        }
        
    }
}
