/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/flow/MemoryPersistenceManager.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.flow;

import static com.netapp.common.flow.FlowConstants.CONTEXT_LIFETIME;

import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import com.netapp.common.MsgKey;
import com.netapp.common.flow.api.OperationState;
import com.netapp.common.logging.LogLevel;
import com.netapp.common.logging.LogMessage;
import com.netapp.common.logging.NALogger;
import com.netapp.common.util.CollectionUtils;
import com.netapp.common.util.ConfigBundle;
import com.netapp.common.util.StringUtils;

/**
 * In-memory persistence manager.
 * This doesn't actually persist, it only works while the process is running.
 * @author sunday
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public class MemoryPersistenceManager implements PersistenceManager {
    private static final NALogger logger = NALogger.getLogger(MemoryPersistenceManager.class);
    
    private OperationLogListener logAppender;
    private Map<String, OperationContext> contexts = new HashMap<String, OperationContext>();
    
    private Map<OperationId, List<LogMessage>> logMessage = new HashMap<OperationId, List<LogMessage>>();
    
    private Map<String, OperationId> locks = new HashMap<String, OperationId>();
    
    private ConfigBundle applicationContext;
    private long contextLifetime = 30;
    
    public MemoryPersistenceManager() {        

    }
    
    public void setApplicationContext(ConfigBundle applicationContext) {
        this.applicationContext = applicationContext;
    }
    
    public void init() {
        if (applicationContext != null) {
            Long value = CONTEXT_LIFETIME.lookup(this.applicationContext);
            this.contextLifetime = (value != null) ? value : this.contextLifetime;
        }
        this.logAppender = new OperationLogListener(this, LogLevel.INFO);
        NALogger.addListener(this.logAppender);
    }
    
    /**
     * 
     * @see com.netapp.common.flow.PersistenceManager#start(Engine)
     */
    public void start(Engine engine) {
    }
    
    /**
     * 
     * @see com.netapp.common.flow.PersistenceManager#stop()
     */
    public void stop() {
    }
    
    /**
     * @see com.netapp.common.flow.PersistenceManager#createContext(com.netapp.common.flow.OperationTemplate, java.util.Map)
     */
    public synchronized OperationContext createContext(OperationTemplate template, Map<String, ?> params) {
        OperationContext context = new OperationContext(this, template, params);
        context.enterContext();
        saveContext(context);
        context.exitContext();
        return context;
    }
    
    public OperationContext getContext(String id) {
        OperationContext result = null;
        result = contexts.get(id);
        return result;
    }

    /**
     * Get the next runnable context for any of the specified operations.
     * As a side effect, as we're traversing the list of operations, we remove any
     * that have expired.
     * 
     * This is a really stupid implementation, and is subject to thrashing under high loads
     * as we only return the first candidate we find.
     * @see com.netapp.common.flow.PersistenceManager#getRunnableContext(java.util.Set)
     */
    public synchronized OperationContext getRunnableContext(Set<String> operationTypes) {
        Date expireAfter = new Date(System.currentTimeMillis() - (this.contextLifetime * 60 * 1000));
        
        Date now = new Date();
        OperationContext result = null;
        Set<OperationContext> expiredContexts = new HashSet<OperationContext>();
        
        for (OperationContext context : this.contexts.values()) {
            if (context.getState().equals(OperationState.PAUSED) 
                    && !context.isDirty()
                    && (context.getHoldUntilTime() == null || context.getHoldUntilTime().before(now))
                    && operationTypes.contains(context.getOperationType())) {
                context.setState(OperationState.RUNNING);
                result = context;
                result.loadWorkingRegisters();
                break;
            }
            else if (context.isTerminalState() && context.getCheckpointTime().before(expireAfter)) {
                expiredContexts.add(context);
            }
        }
        
        for (OperationContext expiredContext : expiredContexts) {
            logger.internal("Expiring completed context " + expiredContext.getId());
            contexts.remove(expiredContext.getId().getId());
            if (!StringUtils.isEmpty(expiredContext.getClientRequestId())) {
                contexts.remove(expiredContext.getClientRequestId());
            }
            logMessage.remove(expiredContext.getId());
        }
        return result;
    }

    /**
     * @see com.netapp.common.flow.PersistenceManager#saveContext(com.netapp.common.flow.OperationContext)
     */
    public synchronized void saveContext(OperationContext context) {
        logger.debug(MsgKey.FLOW_PERSISTENCE_CHECKPOINT, context.getCheckpoint(), context.getState());
        context.markClean();
        contexts.put(context.getId().getId(), context);
        if (!StringUtils.isEmpty(context.getClientRequestId())) {
            contexts.put(context.getClientRequestId(), context);
        }
        if (context.isTerminalState()) {
            releaseLocks(context.getId());
        }
    }

    public synchronized void cancelContext(String id) {
        OperationContext context = this.getContext(id);
        
        if (context == null) {
            return;
        }
        context.enterContext();
        if (!context.isTerminalState()) {
            logger.debug(MsgKey.FLOW_OPERATION_CANCEL_REQUESTED);
            context.cancel();
        }
        else if (context.isAborting()) {
            logger.debug(MsgKey.FLOW_OPERATION_CANNOT_CANCEL_ABORTING_OPERATION);
        }
        else {
            logger.debug(MsgKey.FLOW_OPERATION_CANNOT_CANCEL_FINISHED_OPERATION);
        }
        context.exitContext();
    }
    
    public synchronized void releaseContext(String id) {
        OperationContext context = getContext(id);
        if (context == null) {
            return;
        }
        
        if (!context.isTerminalState()) {
            logger.error(MsgKey.FLOW_PERSISTENCE_CONTEXT_RELEASE_NOT_COMPLETE, context.getId());
            return;
        }
        
        logger.debug(MsgKey.FLOW_PERSISTENCE_CONTEXT_RELEASE, context.getId());
        contexts.remove(context.getId().getId());
        if (!StringUtils.isEmpty(context.getClientRequestId())) {
            contexts.remove(context.getClientRequestId());
        }
        logMessage.remove(context.getId());
    }
    
    public void addMessage(OperationId id, LogMessage message) {
        List<LogMessage> list = null;
        synchronized (logMessage) {
            list = logMessage.get(id);
            if (list == null) {
                list = new ArrayList<LogMessage>();
                logMessage.put(id, list);
            }
        }
        synchronized(list) {
            list.add(message);
        }
    }

    public List<LogMessage> getMessages(OperationId id) {
        return logMessage.get(id);
    }

    public void acquireLocks(OperationId id, Set<String> keys) throws UnableToAcquireLockException {
        OperationId holder = null;
        synchronized (this.locks) {
            for (String key : keys) {
                holder = this.locks.get(key);
    
                // Log outside of the synchronized lock
                if (holder != null && !holder.equals(id)) {
                    OperationContext ctx = getContext(holder.getId());
                    
                    if (ctx != null) {
                        throw new UnableToAcquireLockException(holder, keys,
                            logger.error(MsgKey.FLOW_PERSISTENCE_LOCK_IN_USE, key, 
                                ctx.getOperationType(), 
                                ctx.getId(), 
                                StringUtils.isEmpty(ctx.getClientRequestId()) ? "" : ctx.getClientRequestId()));
                    }
                    else {
                        // this case shouldn't really happen where there is a lock but no operation context
                        // but this is here just as a precaution.
                        logger.warn(MsgKey.FLOW_PERSISTENCE_LOCK_ABANDONED, key, holder.getId());
                        releaseLocks(new OperationId(holder.getId()));
                    }
                }
            }

            for (String key : keys) {
                this.locks.put(key, id);
            }            
        }

        
        logger.debug(MsgKey.FLOW_PERSISTENCE_ACQUIRED_LOCKS, CollectionUtils.collectionToString(keys, ","));
    }

    protected void releaseLocks(OperationId id) {
        logger.internal("Releasing locks held by " + id);
        Set<String> heldLocks = new HashSet<String>();
        
        synchronized (this.locks) {
            Iterator<Map.Entry<String, OperationId>> iter = locks.entrySet().iterator();
            while (iter.hasNext()) {
                Map.Entry<String, OperationId> e = iter.next();
                if (id.equals(e.getValue())) {
                    heldLocks.add(e.getKey());
                    iter.remove();
                }
            }
        }
        
        for (String lock : heldLocks) {
            logger.internal("Released lock %s held by %s", lock, id);
        }
    }
    
    /**
     * Always returns true for this manager
     * @see com.netapp.common.flow.PersistenceManager#isValid()
     */
    public boolean isValid() {
        return true;
    }
    
}
