/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/flow/Operation.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.flow;

import static com.netapp.common.flow.FlowConstants.ABORT_DEBUG_DISABLE;

import java.util.Date;
import java.util.Iterator;

import com.netapp.common.MsgKey;
import com.netapp.common.flow.api.OperationState;
import com.netapp.common.logging.NALogger;
import com.netapp.common.util.Utils;

/**
 * This class executes a hydrated execution context.
 * We reset our state, and then run through the work list
 * in {@link #execute()} until:
 * <ul>
 *   <li>We execute the last ExecutionTask in the current worklist; or
 *   <li>A {@link RetryableException} is caught, causing us to set the context state to paused; or
 *   <li>We catch an exception during an abort to transition to INCOMPLETE.
 * </ul>
 * @author sunday
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
class Operation {
    private static final NALogger logger = NALogger.getLogger(Operation.class);
    private final Engine engine;
    private final OperationContext context;
    Iterator<? extends ExecutionStep> executionPoint;
    int executionStepCount;
    int executionStepsCompleted;
    
    Operation(Engine engine, OperationContext context) {
        assert context.getState() == OperationState.RUNNING;
        this.engine = engine;
        this.context = context;
        this.executionStepCount = this.context.getOperationTemplate().getTaskSteps().size();
    }
    
    public void execute() {
        this.context.enterContext();
        this.engine.lifecycleListenerProxy().operationActivated(this.context.getId(), this.context.getClientRequestId());
        boolean aborting = this.context.isAborting();
        if (!aborting) {
            if (context.getCheckpoint() == null) {
                this.executionPoint = context.getOperationTemplate().getTaskSteps().iterator();
                this.executionStepsCompleted = 0;
            }
            else {
                this.executionPoint = iteratorAtCheckpoint(context.getCheckpoint());  
                this.executionStepsCompleted = this.getExecutionStepsCompletedBeforeCheckpoint(context.getCheckpoint());
            }
            
            try {
                executeCurrentStack();
                this.context.setState(OperationState.COMPLETE);
                this.context.loadPersistentRegisters();
                logger.debug(MsgKey.FLOW_OPERATION_COMPLETE);
            }
            catch (RetryableException e) {
                // Fall through
            }
            // CHECKSTYLE:OFF
            catch (Throwable t) {
            // CHECKSTYLE:ON
                this.context.setTaskException(t);
                this.context.setFailureStrategy(FailureStrategy.IGNORE);
                if (this.context.getAbortCheckpoint() != null) {
                    aborting = true;
                    logger.error(MsgKey.FLOW_OPERATION_ABORTING);
                }
                else {
                    this.context.setState(OperationState.FAILED);
                    logger.error(MsgKey.FLOW_OPERATION_FAILED, t, Utils.prettyString(t));
                }
            }
            this.context.persist();
        }
        
        if (aborting) {
            this.executionPoint = iteratorAtAbort(this.context.getAbortCheckpoint());
            try {
                if (ABORT_DEBUG_DISABLE.lookup(engine.getApplicationContext())) {
                    throw new RuntimeException("Abort processing disabled by property " + ABORT_DEBUG_DISABLE);
                }
                else {
                    executeCurrentStack();
                }
                this.context.setState(OperationState.FAILED);
                logger.warn(MsgKey.FLOW_OPERATION_ABORT_SUCCESSFUL);
                logger.error(MsgKey.FLOW_OPERATION_FAILED, Utils.prettyString(this.context.getTaskException()));
            }           
            catch (RetryableException e) {
                // Fall through
            }
            // CHECKSTYLE:OFF
            catch (Throwable t) {
            // CHECKSTYLE:ON
                this.context.setAbortException(t);
                this.context.setState(OperationState.INCOMPLETE);
                logger.error(MsgKey.FLOW_OPERATION_INCOMPLETE, t, Utils.prettyString(t));
            }
            this.context.persist();
        }
        
        if (!this.context.isTerminalState()) {
            this.engine.lifecycleListenerProxy().operationPaused(this.context.getId(), this.context.getClientRequestId());
        }
        else {
            this.engine.lifecycleListenerProxy().operationCompleted(this.context.getId(), this.context.getClientRequestId());
        }
        this.context.exitContext();
    }

    /**
     * Algorithm for retry backoff.
     * Currently, we start at 1 minute and add an additional minute
     * for each successive retry, capped at 10 minutes (which would be
     * hit after ~ an hour of escalation.)
     * @param retryCount
     * @return
     */
    public static Date getRetryTime(int retryCount) {
        long delayPeriods = Math.min(retryCount, 10);
        return new Date(System.currentTimeMillis() + (1000 * 60 * delayPeriods));
    }
    
    /**
     * 
     */
    private void executeCurrentStack() throws Throwable {
        while (this.executionPoint.hasNext()) {
            ExecutionStep step = this.executionPoint.next();
            
            if (!this.context.isAborting() 
                    && this.context.isCancelled()) {
                throw new CancelException();
            }
            try {
                step.execute(this.context);
                if (!this.context.isAborting() && !this.context.isCancelled()) {
                    this.executionStepsCompleted++;
                    this.engine.lifecycleListenerProxy().operationProgress(
                        this.context.getId(), 
                        this.context.getClientRequestId(),
                        this.executionStepsCompleted,
                        this.executionStepCount);
                }
            }
            catch (RetryableException e) {   
                // Handle a retryable exception by putting the context into a paused state and
                // bubbling the exception up.
                this.context.incrementRetryCount();
                
                if (e.getMaxRetryAttempts() != null && this.context.getRetryCount() > e.getMaxRetryAttempts()) {
                    throw new FlowException(logger.error(MsgKey.FLOW_OPERATION_FAILURE_MAX_RETRIES, e.getMaxRetryAttempts()));
                }
                
                if (e.getHoldUntilTime() != null) {
                    this.context.setHoldUntilTime(e.getHoldUntilTime());
                }
                else {
                    this.context.setHoldUntilTime(getRetryTime(this.context.getRetryCount()));
                }
                this.context.setState(OperationState.PAUSED);
                logger.debug(MsgKey.FLOW_OPERATION_RETRY);
                throw e;
            }
        }
        
        if (this.context.getFailureStrategy() == FailureStrategy.END || this.context.getFailureStrategy() == FailureStrategy.CHECKPOINT) {
            throw new FlowException(logger.debug(MsgKey.FLOW_OPERATION_TASK_LIST_DELAYED_FAILURE));
        }
    }
    
    private Iterator<? extends ExecutionStep> iteratorAtCheckpoint(String checkpoint) {
        Iterator<? extends ExecutionStep> iterator = context.getOperationTemplate().getTaskSteps().iterator();
        while (iterator.hasNext()) {
            ExecutionStep step = iterator.next();
            if ((step instanceof CheckpointTemplate)
                    && ((CheckpointTemplate)step).hasName()
                    && ((CheckpointTemplate)step).getName().equals(checkpoint)) {
                break;
            }
        }
        return iterator;
    }
    
    private int getExecutionStepsCompletedBeforeCheckpoint(String checkpoint) {
        int executionStepsCompletedBeforeCheckpoint = 0;

        for (ExecutionStep step : this.context.getOperationTemplate().getTaskSteps()) {
            ++executionStepsCompletedBeforeCheckpoint;
            if ((step instanceof CheckpointTemplate)
                    && ((CheckpointTemplate)step).hasName()
                    && ((CheckpointTemplate)step).getName().equals(checkpoint)) {
                break;
            }
        }
        
        return executionStepsCompletedBeforeCheckpoint;
    }
    
    private Iterator<? extends ExecutionStep> iteratorAtAbort(String checkpoint) {
        Iterator<? extends ExecutionStep> iterator = context.getOperationTemplate().getAbortSteps().iterator();
        while (iterator.hasNext()) {
            ExecutionStep step = iterator.next();
            if ((step instanceof CheckpointTemplate)
                    && ((CheckpointTemplate)step).hasName()
                    && ((CheckpointTemplate)step).getName().equals(checkpoint)) {
                break;
            }
        }
        return iterator;
    }
    
    @Override
    public String toString() {
        return this.context.toString();
    }
}
