/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/logging/DispersingRollingFileAppender.java#1 $
 * 
 * Copyright (c) 2009 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.common.logging;

import java.io.IOException;
import java.util.Collections;
import java.util.Map;
import java.util.Properties;

import org.apache.commons.collections.map.LRUMap;
import org.apache.log4j.AppenderSkeleton;
import org.apache.log4j.MDC;
import org.apache.log4j.PatternLayout;
import org.apache.log4j.PropertyConfigurator;
import org.apache.log4j.RollingFileAppender;
import org.apache.log4j.spi.LoggingEvent;

/**
 * A log4j {@link org.apache.log4j.Appender} that can be used to disperse log messages to different 
 * log files based on a selector key's value in the {@link MDC}.  
 * 
 * Example <tt>log4j.properties</tt> configuration:
 * <pre>
 * log4j.appender.file=com.netapp.common.logging.DispersingRollingFileAppender
 * log4j.appender.file.threshold=TRACE
 * log4j.appender.file.selector=operationId
 * log4j.appender.file.defaultFile=log/server.log
 * log4j.appender.file.fileFormat=log/op-%X{operationStartDate}-%X{operationName}.log
 * log4j.appender.file.maxBackupIndex=100
 * log4j.appender.file.maxFileSize=500KB
 * log4j.appender.file.layout=org.apache.log4j.PatternLayout
 * log4j.appender.file.layout.ConversionPattern=%d{ISO8601} [%t] [%-5p]: %m%n
 * </pre>
 * 
 * @author sirianni
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public class DispersingRollingFileAppender extends AppenderSkeleton {

    private String maxFileSize;
    
    private int maxBackupIndex = 1;
    
    private Map<String, RollingFileAppender> appenders; 
    
    private String selector;
    
    private PatternLayout fileFormat;

    private int childAppenderCacheSize = 10;
    
    private String defaultFile;
    
    private RollingFileAppender defaultAppender;
    
    @SuppressWarnings("unchecked")
    @Override
    public void activateOptions() {
        /*
         * Use a bounded LRU Map to store the appenders.
         * 
         * Since we don't have a notification handle on when the child appender MDC selector
         * goes out of scope, we simply bound the size of the map.  Child appenders will get
         * recreated as necessary if they happen to get kicked out of the map before the 
         * MDC selector has actually gone out of scope.
         */
        this.appenders = Collections.synchronizedMap(new LRUMap(this.childAppenderCacheSize));
        try {
            this.defaultAppender = new RollingFileAppender(this.layout, this.defaultFile, true);
            this.defaultAppender.setMaxBackupIndex(this.maxBackupIndex);
            this.defaultAppender.setMaxFileSize(this.maxFileSize);
            this.defaultAppender.activateOptions();
        }
        catch (IOException e) {
            System.err.printf("Unable to configure RollingFileAppender for file %s.  Error: %s%n", this.defaultFile, e.getLocalizedMessage());
        }
    }
    
    @Override
    protected void append(LoggingEvent loggingevent) {
        RollingFileAppender appender = this.selectAppender(loggingevent);
        
        if (appender != null) {
            appender.doAppend(loggingevent);    
        }
        else {
            if (this.defaultAppender != null) {
                this.defaultAppender.doAppend(loggingevent);    
            }
        }
    }

    public void close() {
        for (RollingFileAppender appender : this.appenders.values()) {
            appender.close();
        }
    }

    public boolean requiresLayout() {
        return true;
    }
    
    private RollingFileAppender selectAppender(LoggingEvent loggingEvent) {
        String selectorValue = loggingEvent.getProperty(this.selector);
        
        if (selectorValue == null) {
            return null;
        }
        
        RollingFileAppender appender = this.appenders.get(selectorValue);
        
        if (appender == null) {
            
            String file = this.fileFormat.format(loggingEvent);
            
            try {
                appender = new RollingFileAppender(this.getLayout(), file);
                appender.setMaxBackupIndex(this.maxBackupIndex);
                appender.setMaxFileSize(this.maxFileSize);
                appender.activateOptions();
            }
            catch (IOException e) {
                System.err.printf("Unable to configure RollingFileAppender for file %s.  Error: %s%n", file, e.getLocalizedMessage());
            }
            this.appenders.put(selectorValue, appender);
        }
        
        return appender;
    }
    
    /**
     * Set the selector key.  
     * The value of the selector key in the {@link MDC} determines which child appender receives the logging event.
     */
    public void setSelector(String selector) {
        this.selector = selector;
    }

    /**
     * Set the file name format used for child log files.  
     * The file name format follows log4j {@link PatternLayout} syntax and therefore can reference keys in the {@link MDC}.
     * For example: <tt>log/op-%X{operationStartDate}-%X{operationName}.log</tt>
     */
    public void setFileFormat(String fileFormat) {
        this.fileFormat = new PatternLayout(fileFormat);
    }
    
    public void setChildAppenderCacheSize(int childAppenderCacheSize) {
        this.childAppenderCacheSize = childAppenderCacheSize;
    }
    
    public void setDefaultFile(String defaultFile) {
        this.defaultFile = defaultFile;
    }
    
    public void setMaxFileSize(String maxFileSize) {
        this.maxFileSize = maxFileSize;
    }

    public void setMaxBackupIndex(int maxBackupIndex) {
        this.maxBackupIndex = maxBackupIndex;
    }
    
    /*
     * This can't be a unit test because it overrwrites the static log4j configuration
     */
    public static void main(String[] args) {
        
        Properties properties = new Properties();
        
        properties.setProperty("log4j.appender.disperse", "com.netapp.common.logging.DispersingRollingFileAppender");

        properties.setProperty("log4j.appender.disperse.threshold", "DEBUG");
        
        properties.setProperty("log4j.appender.disperse.defaultFile", "log/default.log");
        
        properties.setProperty("log4j.appender.disperse.selector", "testSelector");
        properties.setProperty("log4j.appender.disperse.fileFormat", "log/test-%X{testSelector}.log");
        properties.setProperty("log4j.appender.disperse.childAppenderCacheSize", "2");
        
        properties.setProperty("log4j.appender.disperse.layout", "org.apache.log4j.PatternLayout");
        properties.setProperty("log4j.appender.disperse.layout.ConversionPattern", "%d{ISO8601} [%-5p]: %m%n");
        
        properties.setProperty("log4j.rootLogger", "DEBUG, disperse");
        
        PropertyConfigurator.configure(properties);
        
        NALogger logger = NALogger.getLogger();
        
        MDC.put("testSelector", "foo");
        
        logger.internal("Hello world foo 1");
        
        MDC.put("testSelector", "bar");
        
        logger.internal("Hello world bar");
        
        MDC.put("testSelector", "baz");
        
        logger.internal("Hello world baz");

        MDC.put("testSelector", "foo");
        
        logger.internal("Hello world foo 2");
        
        MDC.remove("testSelector");
        
        logger.internal("Hello world default");
        
    }
    
}
