/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/logging/MessageBundle.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.logging;

import java.util.Arrays;
import java.util.IllegalFormatException;
import java.util.Locale;
import java.util.Map;
import java.util.MissingResourceException;
import java.util.ResourceBundle;
import java.util.WeakHashMap;
import java.util.regex.Matcher;

/**
 * Bundle representing localized message strings.
 * MessageBundles are created by running a list of message keys and text
 * through {@link com.netapp.common.logging.generator.MsgKeyGenerator} to
 * generate enumertations implementing the {@link MessageKey} interface. During runtime,
 * the MessageBundle uses the class of the generated enumerations along with the {@link Bundle}
 * annotation to locate and load the appropriate localized message keys and text.
 * <p>
 * Localized messages may be retrieved via the {@link #localizeMessage(MessageKey)} and {@link #localizeMessage(MessageKey, Object[])}
 * methods.
 * 
 * @author sunday
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public class MessageBundle {
    private static final String DEFAULT_BUNDLE_NAME = "messages";
   
    /** Cache of loaded MessageBundle instances. */
    private static Map<Class<? extends MessageKey>, MessageBundle> instances = new WeakHashMap<Class<? extends MessageKey>, MessageBundle>();
    
    /* Logger for our own use. Note, we can't localize our own messages */
    private static NALogger logger = NALogger.getLogger(MessageBundle.class);
    
    private ResourceBundle bundle;
   
    /**
     * Create a message bundle for the supplied class.
     * Don't call this directly, use {@link #getInstance(MessageKey)} instead
     * to load the correct version from cache.
     * 
     * @param clazz The class of a MsgKey enum
     */
    private MessageBundle(Class<? extends MessageKey> clazz) {
        String bundleName = DEFAULT_BUNDLE_NAME;
        
        Bundle annotation = clazz.getAnnotation(Bundle.class);
        if (annotation != null) {
            bundleName = annotation.value();
        }
        logger.internal("Loading resource bundle from " + bundleName + " for " + clazz.getName());
        
        // TODO: Pick up the correct locale version
        try {
            this.bundle = ResourceBundle.getBundle(bundleName, Locale.getDefault(), clazz.getClassLoader());
        }
        catch (MissingResourceException e) {
            throw new LoggingException("Error loading resource bundle " + bundleName + " for class " + clazz, e);
        }
    }
   
    /**
     * Return the correct instance containing the resource bundle the
     * supplied MsgKey was generated from, loading it if required.
     * @param key MsgKey enumeration to retrieve
     * @return Instance of the message bundle for the supplied key
     */
    static MessageBundle getInstance(MessageKey key) {
        Class<? extends MessageKey> clazz = key.getClass();
        MessageBundle ret = instances.get(clazz);
        if (ret == null) {
            ret = new MessageBundle(clazz);
            instances.put(clazz, ret);
        }
        return ret;
    }
    
    /**
     * Localize the supplied message key from the current resource bundle
     * @param key MsgKey enum to lookup
     * @return Localized text string
     */
    public static String localizeMessage(MessageKey key) {
        return getInstance(key).bundle.getString(key.getPropertyName());
    }
   
    /**
     * Localize the supplied message key from the current resource bundle
     * @param key MsgKey enum to lookup
     * @param params Objects for variable substitution in string
     * @return Localized text string
     */
    public static String localizeMessage(MessageKey key, Object... params) {
        String result = null;
        try {
            result = String.format(localizeMessage(key), params);
        }
        catch (IllegalFormatException e) {
            String args = Arrays.deepToString(params);
            result = key.toString() + " " + args;
            
            logger.internal("Exception while formatting " 
                + key.toString() + " "
                + localizeMessage(key).replaceAll("%", "%%") + " "
                + args + " : "
                + e.getLocalizedMessage(),
                e);
        }
        return result;
    }
    
    /**
     * Extract the APPL-01234 code from a message key 
     * @param key
     * @return The "APPL-01234" code for the supplied message key, or null
     * if the requested message key is not coded.
     */
    public static String code(MessageKey key) {
        return code(localizeMessage(key));
    }
    
    /**
     * Extract the AAPL-01234 code from a string
     * @param msg
     * @return The "APPL-10234" code for the supplied string, or null if
     * msg does not start with a code.
     */
    public static String code(String msg) {
        Matcher m = MessageKey.CODE_PATTERN.matcher(msg);
        if (m.matches()) {
            return m.group(1);
        }
        return null;
    }
    
}
