/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/platform/hpux/HPUXDeviceAnalyzer.java#1 $
 *
 * Copyright (c) 2006 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.platform.hpux;

import java.io.File;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.netapp.common.MsgKey;
import com.netapp.common.logging.NALogger;
import com.netapp.common.platform.DeviceAnalyzer;
import com.netapp.common.platform.DeviceType;
import com.netapp.common.platform.PlatformException;
import com.netapp.common.platform.RawDeviceMapping;

/**
 * Provide a simple mapping between block devices and raw on HP-UX systems.
 *
 * @author sunday
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
class HPUXDeviceAnalyzer implements DeviceAnalyzer {
    
    private static final NALogger logger = NALogger.getLogger(HPUXDeviceAnalyzer.class);

    /** 
     * Pattern to match raw devices. 
     * 11v3 added /dev/disk and /dev/rdisk for MPIO devices, so we look for both
     * /dev/rdsk and /dev/rdisk and map appropriately.
     *
     */
    protected static final Pattern RAW_PATTERN = Pattern.compile("/dev/r(di?sk)/(.*)");

    /** Pattern to match block devices. */
    protected static final Pattern BLOCK_PATTERN = Pattern.compile("/dev/(di?sk)/(.*)");

    public HPUXDeviceAnalyzer() {
        super();
    }

    /**
     * Map a raw device to the block device. 
     * /dev/rdsk/c1t0d0 ==&gt; /dev/dsk/c1t0d0
     *
     * @param rawDevice The raw device path to map.
     * @return The associate block device
     */
    public RawDeviceMapping lookupMappingByRawDevice(File rawDevice) {
        Matcher m = RAW_PATTERN.matcher(rawDevice.getAbsolutePath());
        if (m.matches()) {
            File disk = new File("/dev/" + m.group(1) + "/" + m.group(2));
            return new RawDeviceMapping(rawDevice, disk);
        }
        else {
            String error = logger.error(MsgKey.PLATFORM_RAW_DEVICE_MAPPER_UNSUPPORTED_OPERATION);
            throw new PlatformException(error);
        }
    }

    /**
     * Map a disk to the corresponding raw device.
     *
     * @param disk Disk entry to map
     * @return Raw device name
     */
    public RawDeviceMapping lookupMappingByDisk(File disk) {
        Matcher m = BLOCK_PATTERN.matcher(disk.getAbsolutePath());
        if (m.matches()) {
            File rawDevice = new File("/dev/r" + m.group(1) + "/" + m.group(2));
            return new RawDeviceMapping(rawDevice, disk);
        }
        else {
            String error = logger.error(MsgKey.PLATFORM_RAW_DEVICE_MAPPER_UNSUPPORTED_OPERATION);
            throw new PlatformException(error);
        }
    }

    /**
     * Disk and partition are the same on HP-UX. Partitioning is done via
     * VxVM which we don't support.
     * 
     */
    public String extractPartitionPathElement(File path) {
        return "";
    }

    public int extractPartitionNumber(File path) {
        throw new UnsupportedOperationException();
    }
    
    /**
     * The raw device is always the whole disk.
     * 
     */
    public String extractFullDiskPathElement(File path) {
        return path.getAbsolutePath();
    }

    public File deviceForPartition(File disk, int partitionNumber) {
        String partition = Integer.toString(partitionNumber);
        return this.deviceForPartition(disk, partition);
    }
    
    public File deviceForPartition(File disk, String partition) {
        return new File(disk + partition);
    }

    public DeviceType determineDeviceType(File file) {
        if (RAW_PATTERN.matcher(file.getAbsolutePath()).matches()) {
            return DeviceType.RAW_DEVICE;
        }
        else if (BLOCK_PATTERN.matcher(file.getAbsolutePath()).matches()) {
            return DeviceType.BASE_DEVICE;
        }
        else {
            return DeviceType.NON_DEVICE;    
        }
    }
}
