/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/platform/linux/RawDevicesFile.java#1 $
 * 
 * Copyright (c) 2006 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.platform.linux;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.List;

import com.netapp.common.MsgKey;
import com.netapp.common.flow.OperationContext;
import com.netapp.common.flow.TaskUtils;
import com.netapp.common.logging.MessageBundle;
import com.netapp.common.logging.NALogger;
import com.netapp.common.platform.PlatformException;
import com.netapp.common.platform.RawDeviceMapping;
import com.netapp.common.util.Utils;

public abstract class RawDevicesFile {
    private static final NALogger logger = NALogger.getLogger();

    private static final String COMMENT_PREFIX = "# ";
    
    /* Raw devices comment from 1.1.1 */
    private static final String COMMENT_1_1_1 = "^# \\[[0-9a-f]{32}\\] SMO .*";

    protected abstract File getFile();
    
    protected abstract RawDeviceMapping mappingFromEntry(String entry);
    
    protected abstract String mappingToEntry(RawDeviceMapping mapping);
    
    public List<RawDeviceMapping> getRawDeviceMappings() {
        List<RawDeviceMapping> mappings = new ArrayList<RawDeviceMapping>();
        try {
            BufferedReader bufferedReader = new BufferedReader(new FileReader(this.getFile()));
            for (String line = bufferedReader.readLine(); line != null; line = bufferedReader.readLine()) {
                RawDeviceMapping mapping = this.mappingFromEntry(line);
                if (mapping != null) {
                    mappings.add(mapping);
                }
            }
        }
        catch (FileNotFoundException e) {
            // Ignore - it's not an error if the file doesn't exist.
            logger.internal("Raw devices file \"%s\" does not exist", this.getFile().getAbsolutePath());
        }
        catch (IOException e) {
            String error = logger.error(MsgKey.PLATFORM_UNABLE_TO_LOAD_RAW_DEVICE, 
                e, this.getFile().getAbsolutePath(), Utils.prettyString(e));
            throw new PlatformException(error, e);
        }
        return mappings;
    }

    public void addMapping(RawDeviceMapping mapping) {
        try {
            List<String> output = new ArrayList<String>();
            
            try {
                BufferedReader bufferedReader = new BufferedReader(new FileReader(this.getFile()));

                // Read in the file 
                for (String line = bufferedReader.readLine(); line != null; line = bufferedReader.readLine()) {
                    output.add(line);
                }
                bufferedReader.close();
            }
            catch (FileNotFoundException e) {
                // Ignore
                logger.internal("Raw devices file \"%s\" does not exist; will create a new one", this.getFile().getAbsolutePath());
            }
            
            // Now overwrite the file
            PrintWriter printWriter = new PrintWriter(new BufferedWriter(new FileWriter(this.getFile(), false)));
            for (String line : output) {            
                printWriter.println(line);
            }
            
            // Add the mapping
            String comment = null;
            if (OperationContext.getThreadContext() != null) {
                comment = MessageBundle.localizeMessage(MsgKey.PLATFORM_RAW_DEVICE_MAPPER_FILE_COMMENT, new Object[] { TaskUtils.currentOperationId() });
            }
            else {
                comment = MessageBundle.localizeMessage(MsgKey.PLATFORM_RAW_DEVICE_MAPPER_FILE_COMMENT_NO_OPERATION);                
            }
            printWriter.println(COMMENT_PREFIX + comment);
            printWriter.println(this.mappingToEntry(mapping));
            printWriter.close();
        }
        catch (IOException e) {
            String error = logger.error(MsgKey.PLATFORM_RAW_DEVICE_MAPPER_CANNOT_WRITE_RAW_DEVICE_FILE, 
                this.getFile().getAbsolutePath(), Utils.prettyString(e));
            throw new PlatformException(error, e);
        }
    }
    
    public void removeMapping(RawDeviceMapping rawDeviceMapping) {
        // get the first 20 characters of our comment in case it is ever localized, just use the no operation one since 
        // it's basically the same message.
        
        final String commentStart = MessageBundle.localizeMessage(MsgKey.PLATFORM_RAW_DEVICE_MAPPER_FILE_COMMENT_NO_OPERATION).substring(0, 20);
        
        try {
            List<String> output = new ArrayList<String>();
            BufferedReader bufferedReader = new BufferedReader(new FileReader(this.getFile()));
            for (String line = bufferedReader.readLine(); line != null; line = bufferedReader.readLine()) {
                RawDeviceMapping mappingInFile = this.mappingFromEntry(line);
                if (mappingInFile != null && mappingInFile.getRawDevice().equals(rawDeviceMapping.getRawDevice())) {
                    String previousLine = output.get(output.size() - 1);
                    if (previousLine.startsWith(COMMENT_PREFIX + commentStart)
                            || previousLine.matches(COMMENT_1_1_1)) {
                        // Remove
                        output.remove(output.size() - 1);
                    }
                    else {
                        String error = logger.error(MsgKey.PLATFORM_RAW_DEVICE_MAPPER_NOT_NETAPP_GENERATED,
                            line);
                        throw new PlatformException(error);
                    }
                }
                else {
                    output.add(line);
                }
            }
            bufferedReader.close();
            
            // Now overwrite the file
            PrintWriter printWriter = new PrintWriter(new BufferedWriter(new FileWriter(this.getFile(), false)));
            for (String line : output) {            
                printWriter.println(line);
            }
            printWriter.close();
        }
        catch (FileNotFoundException e) {
            // Ignore - If the file doesn't exist our mapping is already gone
            logger.internal("Raw devices file \"%s\" does not exist", this.getFile().getAbsolutePath());
            
        }
        catch (IOException e) {
            String error = logger.error(MsgKey.PLATFORM_RAW_DEVICE_MAPPER_CANNOT_WRITE_RAW_DEVICE_FILE, 
                this.getFile().getAbsolutePath(), Utils.prettyString(e));
            throw new PlatformException(error, e);
        }
    }
}
