/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/server/AbstractJettyServer.java#1 $
 * 
 * Copyright (c) 2009 NetApp, Inc. 
 * All rights reserved.
 */
package com.netapp.common.server;

import static com.netapp.common.util.ConfigProperty.booleanProperty;
import static com.netapp.common.util.ConfigProperty.fileProperty;
import static com.netapp.common.util.ConfigProperty.intProperty;
import static com.netapp.common.util.ConfigProperty.stringProperty;

import java.io.File;
import java.util.List;

import org.mortbay.jetty.Connector;
import org.mortbay.jetty.nio.SelectChannelConnector;
import org.mortbay.jetty.security.SslSocketConnector;

import com.netapp.common.MsgKey;
import com.netapp.common.logging.MessageBundle;
import com.netapp.common.logging.NALogger;
import com.netapp.common.logging.ReloadLoggingSignalHandlerImpl;
import com.netapp.common.platform.OperatingSystem;
import com.netapp.common.signal.SignalHandlerImpl;
import com.netapp.common.util.ConfigBundle;
import com.netapp.common.util.ConfigProperty;
import com.netapp.common.util.Pid;

/**
 * Provides default behavior for starting a jetty server.  Sub-classes are responsible for instantiating the actual
 * <b>org.mortbay.jetty.Server</b> and <b>ConfigBundle</b>.
 * 
 * @author claird
 * 
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public abstract class AbstractJettyServer implements Server {

    private static final NALogger LOGGER = NALogger.getLogger(AbstractJettyServer.class);
    
    public static final ConfigProperty<Integer> PORT_PROPERTY = intProperty("http.port", 8080);
    public static final ConfigProperty<Boolean> SSL_ENABLE_PROPERTY = booleanProperty("http.ssl.enable", false);
    public static final ConfigProperty<File> SSL_KEYSTORE_FILE_PROPERTY = fileProperty("http.ssl.keystore.file", new File("etc/keystore"));
    public static final ConfigProperty<String> SSL_KEYSTORE_PASSWORD_PROPERTY = stringProperty("http.ssl.keystore.password", "");
    public static final ConfigProperty<String> SSL_KEY_PASSWORD_PROPERTY = stringProperty("http.ssl.key.password", "");
    public static final String DEFAULT_BEAN_NAME = "jetty.server";
    public static final String DEFAULT_SERVER_NAME = "jetty";
    public static final String DEFAULT_CONFIG_BUNDLE_NAME = "config.bundle";
    
    private String beanName;
    private String serverName;
    private org.mortbay.jetty.Server server;
    private ConfigBundle configBundle;
    private ServerDescriptor descriptor;
    private int port = -1;
    
    /**
     * 
     */
    public AbstractJettyServer() {
        super();
    }

    public String getBeanName() {
        return beanName;
    }

    public void setBeanName(String beanName) {
        this.beanName = beanName;
    }
    
    public String getServerName() {
        return serverName;
    }

    /**
     * This is optional to set. If not set, the default value of 'jetty' will be used.
     * @param serverName
     */
    public void setServerName(String serverName) {
        this.serverName = serverName;
    }

    public org.mortbay.jetty.Server getServer() {
        return server;
    }

    public void setServer(org.mortbay.jetty.Server server) {
        this.server = server;
    }
    
    public ConfigBundle getConfigBundle() {
        return configBundle;
    }

    public void setConfigBundle(ConfigBundle configBundle) {
        this.configBundle = configBundle;
    }

    public ServerDescriptor getDescriptor() {
        return descriptor;
    }

    public void setDescriptor(ServerDescriptor descriptor) {
        this.descriptor = descriptor;
    }
    
    public int getPort() {
        return this.port;
    }

    public void setPort(int port) {
        this.port = port;
    }

    public void start() {
        loadConfiguration();
        Pid pid = new Pid();
        
        registerSignalHandlers();
        
        try {
            getServer().start();
            
            while (!getServer().isStarted()) {
                Thread.sleep(1000);
            }
        }
        /*
         * CHECKSTYLE:OFF
         * The Jetty code really does just through an exception
         */
        catch (Exception e) {
            LOGGER.error(MsgKey.SERVER_JETTY_SERVER_FAILED, this.port);
            throw new ServerException(e);
        }
        /*
         * CHECKSTYLE:ON
         */
        
        LOGGER.info(MsgKey.SERVER_JETTY_SERVER_STARTED, this.port);
        setDescriptor(new ServerDescriptor(getServerName(), pid.getValue(), this.port));
    }

    public void stop() {
        try {
            if (null != getServer()) {
                getServer().stop();
            }
            else { // stop it by loading the pid file first then issuing a kill command
                loadConfiguration();
                List<ServerDescriptor> descriptors = getDescriptor().getAllDescriptors();
                ServerDescriptor foundDescriptor = null;
                
                for (ServerDescriptor descrip : descriptors) {
                    if (descrip.getPort() == port) {
                        foundDescriptor = descrip;
                        break;
                    }
                }
                
                if (null != foundDescriptor) {
                    LOGGER.info(MsgKey.SERVER_JETTY_SERVER_STOPPING, getDescriptor().getPort());
                    OperatingSystem.createInstance().kill(foundDescriptor.getPid(), 9);
                    /*
                    String cmd = "kill -9 " + String.valueOf(foundDescriptor.getPid());
                    Process process = Runtime.getRuntime().exec(cmd);
                    
                    process.waitFor();
                    int exitValue = process.exitValue();
                    process.destroy();
                    
                    MessageKey key = (0 == exitValue) ? MsgKey.SERVER_JETTY_SUCCESSFUL_EXEC_COMMAND : MsgKey.SERVER_JETTY_FAILURE_EXEC_COMMAND;
                    LOGGER.info(key, cmd, exitValue);  
                    */                  

                    foundDescriptor.deleteFile();                    

                    /*
                    if (exitValue != 0) {
                        throw new ServerException(LOGGER.error(MsgKey.SERVER_JETTY_FAILURE_EXEC_COMMAND, cmd, exitValue));
                    }
                    */
                }
                else {
                    throw new ServerException(LOGGER.error(MsgKey.SERVER_JETTY_INVALID_PID));
                }
            }
            
            LOGGER.info(MsgKey.SERVER_JETTY_SERVER_STOPPED_ON_PORT, getDescriptor().getPort());
            
            while (!getServer().isStopped()) {
                Thread.sleep(1000);
            }

        }
        /*
         * CHECKSTYLE:OFF
         * The Jetty code really does just through an exception
         */
        catch (Exception e) {
            throw new ServerException(MessageBundle.localizeMessage(MsgKey.SERVER_JETTY_SERVER_STOP_FAILED, e.getMessage()));
        }
        /*
         * CHECKSTYLE:ON
         */
    }

    protected void registerSignalHandlers() {
        /*
         * Don't let the Jetty server die on INTs, HUPs, and TERMs
         * The way to kill it is to send a 'stop' SOAP message 
         * Some VMs may not let us do this, in which case SignalHandlerImpl.replace
         * will log a debug
         */

        new SignalHandlerImpl("INT", true, SignalHandlerImpl.ProgramType.SERVER).replace();
        new SignalHandlerImpl("TERM", false, SignalHandlerImpl.ProgramType.SERVER).replace();
        new ReloadLoggingSignalHandlerImpl("HUP", false, SignalHandlerImpl.ProgramType.SERVER).replace();
    }
    
    /**
     * Retrieve the port on which the server is/will be listening based on the configuration.
     * 
     * @return TCP Port server is configured to listen on
     */
    public int getServerPort() {
        return this.port;
    }

    protected void loadConfiguration() {
        this.loadComponents();
        
        if (this.port == -1) {
            this.port = PORT_PROPERTY.lookup(getConfigBundle());
        }
        Connector connector = null;
        if (SSL_ENABLE_PROPERTY.lookup(getConfigBundle())) {
            SslSocketConnector c = new SslSocketConnector();
            c.setKeystore(SSL_KEYSTORE_FILE_PROPERTY.lookup(getConfigBundle()).getAbsolutePath());
            c.setPassword(SSL_KEYSTORE_PASSWORD_PROPERTY.lookup(getConfigBundle()));
            c.setKeyPassword(SSL_KEY_PASSWORD_PROPERTY.lookup(getConfigBundle()));
            c.setProtocol("SSL");
            connector = c;
        }
        else {
            connector = new SelectChannelConnector();
        }
        connector.setPort(this.port);
        this.server.addConnector(connector);
    }

    /**
     * Load the jetty server and config bundle components.
     */
    protected abstract void loadComponents();
}