/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/spring/lazilyautowire/LazilyAutowiredInvocationHandler.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.spring.lazilyautowire;

import java.lang.reflect.Field;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.Set;

import net.sf.cglib.proxy.MethodInterceptor;
import net.sf.cglib.proxy.MethodProxy;

import org.springframework.beans.FatalBeanException;
import org.springframework.beans.TypeConverter;
import org.springframework.beans.factory.config.ConfigurableListableBeanFactory;
import org.springframework.beans.factory.config.DependencyDescriptor;
import org.springframework.util.ReflectionUtils;

import com.netapp.common.logging.NALogger;

/**
 * InvocationHandler for proxies of LazilyAutowired components.
 * 
 * The first time a method is invoked on the LazilyAutowired object, the
 * target bean will be retrieved from the beanFactory and cached.
 * 
 * Then, the LazilyAutowiredInvocationHandler will attempt to
 * replace the proxy in the parent bean with the target bean.
 * 
 * If the replacement is successful, then the proxy will no longer be
 * referenced and can be garbage collected.
 * 
 * If the replacement is not successful, then each additional time a method
 * is invoked, the proxy will delegate to the cached target bean. 
 * 
 * Works for both JDK proxies and CGLIB proxies.
 * 
 * @author philipc
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public class LazilyAutowiredInvocationHandler implements InvocationHandler, MethodInterceptor {
    
    private static final NALogger logger = NALogger.getLogger();

    private final ConfigurableListableBeanFactory beanFactory;
    private final String parentBeanName;
    private final DependencyDescriptor descriptor;
    private final Object parentBean; 
    
    private Object target;
    private boolean canReplaceProxyWithTarget;

    /**
     * @param beanFactory the beanFactory from which to resolve the target bean
     * @param parentBeanName the name of the bean that needs a reference to the target bean
     * @param parentBean (optional) the bean that has a reference to the proxy (if null, the proxy will not be replaced with the reference to the target bean)  
     * @param descriptor the descriptor to resolve to find the target bean
     */
    public LazilyAutowiredInvocationHandler(ConfigurableListableBeanFactory beanFactory, String parentBeanName, Object parentBean, DependencyDescriptor descriptor) {
        this.beanFactory = beanFactory;
        this.parentBeanName = parentBeanName;
        this.descriptor = descriptor;
        this.parentBean = parentBean;
        this.canReplaceProxyWithTarget = (parentBean != null);
    }

    /*
     * The JDK proxy start point
     */
    public Object invoke(Object proxy, Method method, Object[] args) throws Throwable {
        return executeMethodCall(method, args);
    }

    /*
     * The CGLIB proxy start point
     */
    public Object intercept(Object object, Method adapterMethod, Object[] args, MethodProxy methodProxy) throws Throwable {
        return executeMethodCall(adapterMethod, args);
    }
    
    private Object executeMethodCall(Method method, Object[] args) throws Throwable {
        
        if ("finalize".equals(method.getName()) && (args == null || args.length == 0)) {
            // burt 298616:  Don't pass along the finalize method call.
            return null;
        }
        
        if (this.target == null) {
            this.target = resolveTarget();
            
            replaceProxyWithTarget();
        }
        try {
            Method realMethod = this.target.getClass().getMethod(method.getName(), method.getParameterTypes());
        
            ReflectionUtils.makeAccessible(realMethod);
            return realMethod.invoke(this.target, args);
        }
        catch (InvocationTargetException e) {
            throw e.getCause();
        }
    }
    
    /**
     * Looks the target bean up in the bean factory
     */
    private Object resolveTarget() {
        
        try {
            Set<String> autowiredBeanNames = new LinkedHashSet<String>(4);
            TypeConverter typeConverter = this.beanFactory.getTypeConverter();
            Object myTarget = this.beanFactory.resolveDependency(
                this.descriptor,
                this.parentBeanName, autowiredBeanNames, typeConverter);
            registerDependentBeans(this.parentBeanName, autowiredBeanNames);
            
            return myTarget;
        }
        // CHECKSTYLE:OFF
        catch (Throwable ex) {
            // CHECKSTYLE:ON
            if (isField()) {
                throw new FatalBeanException("Could not lazily autowire field: " + this.descriptor.getField(), ex);
            }
            else {
                throw new FatalBeanException("Could not lazily autowire method: " + this.descriptor.getMethodParameter().getMethod() + " parameter: " + this.descriptor.getMethodParameter().getParameterIndex(), ex);
            }
        }
    }
    
    private void registerDependentBeans(String beanName, Set<String> autowiredBeanNames) {
        for (Iterator it = autowiredBeanNames.iterator(); it.hasNext();) {
            String autowiredBeanName = (String) it.next();
            this.beanFactory.registerDependentBean(autowiredBeanName, beanName);
            logger.internal("Lazily autowiring by type from bean name '" + beanName + "' via "
                + (isField() ? "field" : "configuration method") + " to bean named '" + autowiredBeanName + "'");
        }
    }

    /**
     * Tries to replace the reference to the proxy object with the target in the parent bean.
     */
    private void replaceProxyWithTarget() {
        if (this.canReplaceProxyWithTarget) {
            if (isField()) {
                try {
                    Field field = this.descriptor.getField();
                    ReflectionUtils.makeAccessible(field);
                    field.set(this.parentBean, this.target);
                    logger.internal("Replaced proxy with lazily autowired bean in %s for field %s", this.parentBeanName, field);
                }
                // CHECKSTYLE:OFF
                catch (Throwable ex) {
                    // CHECKSTYLE:ON
                    this.canReplaceProxyWithTarget = false;
                }
            }
            else {
                Method method = this.descriptor.getMethodParameter().getMethod();
                if (method.getParameterTypes().length == 1) {
                    try {
                        ReflectionUtils.makeAccessible(method);
                        method.invoke(this.parentBean, this.target);
                        logger.internal("Replaced proxy with lazily autowired bean in %s for method %s", this.parentBeanName, method.getName());
                    }
                    // CHECKSTYLE:OFF
                    catch (Throwable ex) {
                        // CHECKSTYLE:ON
                        this.canReplaceProxyWithTarget = false;
                    }
                }
                else {
                    this.canReplaceProxyWithTarget = false;
                }
            }
        }
    }

    private boolean isField() {
        return this.descriptor.getField() != null;
    }

}
