/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/util/CollectionUtils.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.util;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.lang.SystemUtils;

/**
 * Utilities for working with classes in the java collections framework.
 * @author sirianni
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public class CollectionUtils {
    
    protected CollectionUtils() {
    }
    
    /**
     * Returns true if the collection is null or empty.
     * @param collection The collection to check for empty
     * @return true if null or empty; false otherwise
     */
    public static boolean isEmpty(Collection<?> collection) {
        return collection == null || collection.isEmpty();
    }
    
    /**
     * Determines if the supplied Map instance is empty.
     * @param m The Map instance to check for emptiness
     * @return Returns <code>true</code> if null else <code>m.isEmpty()</code>
     */
    public static boolean isEmpty(Map<?, ?> m) {
        return (m == null) || m.isEmpty();
    }

    /**
     * Determines if the supplied array instance is empty
     * @param a The array instance to check for emptiness
     * @return <code>true</code> if null or <code>a.length == 0</code>
     */
    public static boolean isEmpty(Object[] a) {
        return a == null || a.length == 0;
    }
    
    public static <R, S extends R, T extends Collection<R>> T addAll(T collection, S... ss) {
        for (S s : ss) {
            collection.add(s);
        }
        return collection;
    }
    
    public static <R, S, T extends Collection<S>> T transform(Collection<? extends R> oldCollection, T newCollection, Transformer<R, S> transformer) {
        for (R r : oldCollection) {
            newCollection.add(transformer.transform(r));
        }
        return newCollection;
    }

    public static <T> List<? extends List<T>> partition(List<T> list, int bucketSize) {
        LinkedList<LinkedList<T>> buckets = new LinkedList<LinkedList<T>>();
        
        LinkedList<T> currentBucket = new LinkedList<T>();
        buckets.add(currentBucket);
        
        for (T t : list) {
            if (currentBucket.size() == bucketSize) {
                currentBucket = new LinkedList<T>();
                buckets.add(currentBucket);
            }
            currentBucket.add(t);
        }
        
        return buckets;
    }
    
    public static String collectionToString(Collection<?> c, String delimiter) {
        StringBuilder stringBuilder = new StringBuilder();
        for (Iterator<?> i = c.iterator(); i.hasNext();) {
            Object o = i.next();
            stringBuilder.append(o.toString());
            if (i.hasNext()) {
                stringBuilder.append(delimiter);
            }
        }
        return stringBuilder.toString();
    }
    
    public static String collectionToString(Collection<?> c, String delimiter, String prefix, String suffix) {
        StringBuilder stringBuilder = new StringBuilder();
        for (Iterator<?> i = c.iterator(); i.hasNext();) {
            Object o = i.next();
            stringBuilder.append(prefix);
            stringBuilder.append(o.toString());
            stringBuilder.append(suffix);
            if (i.hasNext()) {
                stringBuilder.append(delimiter);
            }
        }
        return stringBuilder.toString();
    }
    
    /**
     * Takes a list of strings and concatenates them into one large string separated
     * by system-dependent newline characters.
     */
    public static String collectionToString(Collection<?> c) {
        return collectionToString(c, SystemUtils.LINE_SEPARATOR);
    }
    
    /**
     * Returns a new Set containing the union of the collections.
     * Duplicates are not preserved
     * @param collections The set of collections to union
     * @return A new Set containing the union of the collections.
     */
    public static <E> Set<E> union(Collection<E>... collections) {
        return unionInto(new HashSet<E>(), collections);
    }

    /**
     * Union the collections into the Collection provided
     * Duplicates are/are not preserved according to the semantics of the provided Collection
     * @param union Collection to hold the union result
     * @param collections The set of collections to union
     * @return The input Collection, now containing the union of the other collections.
     */
    public static <E, S extends Collection<E>> S unionInto(S union, Collection<E>... collections) {
        for (Collection<E> collection : collections) {
            union.addAll(collection);    
        }

        return union;
    }
    
    /**
     * Returns a new Set containing the union of all the collections
     * Duplicates are not preserved
     */
    public static <E> Set<E> flatten(Collection<? extends Collection<E>> collections) {
        Set<E> set = new HashSet<E>();
        for (Collection<E> collection : collections) {
            set.addAll(collection);
        }
        return set;
    }
    
    public static <E> Set<E> dedupe(Collection<E> collections) {
        Set<E> set = new HashSet<E>();
        for (E e : collections) {
            set.add(e);
        }
        return set;
    }
    
    /**
     * Returns a new Set containing the intersection of the two sets.
     * @param s1 The first Set 
     * @param s2 The second Set to intersect it with
     * @return A new Set containing the intersection of the two sets.
     */
    public static <E> Set<E> intersect(Set<E> s1, Set<E> s2) {
        Set<E> intersectSet = new HashSet<E>();
        for (E o : s1) {
            if (s2.contains(o)) {
                intersectSet.add(o);
            }
        }
        return intersectSet;
    }
    
    public static <E> boolean containsAny(Set<E> s1, Set<E> s2) {
        return !intersect(s1, s2).isEmpty();
    }
    
    public static <E> Set<E> difference(Set<E> s1, Set<E> s2) {
        Set<E> result = new HashSet<E>();
        for (E e : s1) {
            if (!s2.contains(e)) {
                result.add(e);    
            }
        }
        return result;
    }
    
    public static <T> T[] append(T[] tArray, T... tsToAppend) {
        List<T> retVal = new ArrayList<T>(tArray.length + tsToAppend.length);
        for (T t : tArray) {
            retVal.add(t);
        }
        for (T t : tsToAppend) {
            retVal.add(t);
        }
        return retVal.toArray(tArray);
    }
    
    public static <T> T[] prepend(T[] tArray, T... tsToPrepend) {
        List<T> retVal = new ArrayList<T>(tArray.length + tsToPrepend.length);
        for (T t : tsToPrepend) {
            retVal.add(t);
        }
        for (T t : tArray) {
            retVal.add(t);
        }
        return retVal.toArray(tArray);
    }    

    public static <S, T> Map<S, T> union(Map<S, T>... maps) {
        Map<S, T> result = new HashMap<S, T>();
        
        for (Map<S, T> map : maps) {
            result.putAll(map);
        }
        
        return result;
    }
    
    public static <R, S, T> Map<R, T> merge(Map<R, S> map1, Map<S, T> map2) {
        Map<R, T> result = new HashMap<R, T>();
        
        for (Map.Entry<R, S> entry : map1.entrySet()) {
            result.put(entry.getKey(), map2.get(entry.getValue()));
        }
        
        return result;
    }
    
    public static <R, S, T, U> Map<R, U> merge(Map<R, S> map1, Map<T, U> map2, Transformer<S, T> transformer) {
        Map<R, U> result = new HashMap<R, U>();
        
        for (Map.Entry<R, S> entry : map1.entrySet()) {
            result.put(entry.getKey(), map2.get(transformer.transform(entry.getValue())));
        }
        
        return result;
    }
    
    public static <S, T> Set<T> getAll(Set<S> keys, Map<S, T> map) {
        Set<T> values = new HashSet<T>();
        for (S key : keys) {
            values.add(map.get(key));
        }
        return values;
    }
    
    public static <R, S, T, U> Map<T, U> transform(Map<R, S> map, Transformer<R, T> transformer1, Transformer<S, U> transformer2) {
        Map<T, U> result = new HashMap<T, U>();
        
        for (Map.Entry<R, S> entry : map.entrySet()) {
            result.put(transformer1.transform(entry.getKey()), transformer2.transform(entry.getValue()));
        }
        
        return result;
    }
    
    public static <S, T> Map<S, T> map(Collection<S> collection, Transformer<? super S, T> transformer) {
        Map<S, T> result = new HashMap<S, T>();
        
        for (S s : collection) {
            result.put(s, transformer.transform(s));
        }
        
        return result;
    }
    
    /**
     * Index a collection by a key
     * @param <S> The collection type
     * @param <T> The key type
     * @param collection The collection to be indexed
     * @param transformer A transformer that extracts the key from the collection element
     */
    public static <S, T> Map<T, S> index(Collection<S> collection, Transformer<? super S, T> transformer) {
        Map<T, S> result = new HashMap<T, S>();
        
        for (S s : collection) {
            result.put(transformer.transform(s), s);
        }
        
        return result;
    }
    
    public static <S, T> Map<T, Set<S>> reverse(Map<S, T> map) {
        Map<T, Set<S>> result = new HashMap<T, Set<S>>();
        
        for (Map.Entry<S, T> entry : map.entrySet()) {
            Set<S> ss = result.get(entry.getValue());
            if (ss == null) {
                ss = new HashSet<S>();
                result.put(entry.getValue(), ss);
            }
            ss.add(entry.getKey());
        }
        
        return result;
    }
    
    public static <T> T getSingleMember(Collection<T> collection) {
        assert collection.size() == 1;
        return collection.iterator().next();
    }
    
    public static <T> T getIgnoreCase(Map<String, T> map, String key) {
        for (Map.Entry<String, T> entry : map.entrySet()) {
            if (entry.getKey().equalsIgnoreCase(key)) {
                return entry.getValue();
            }
        }
        return null;
    }
    
    /**
     * @return The empty Set if the input set is null, else the input set.
     */
    public static <T> Set<T> nullSafe(Set<T> set) {
        Set<T> emptySet = Collections.emptySet();
        
        return set == null ? emptySet : set;
    }

    /**
     * @return The empty List if the input list is null, else the input list.
     */
    public static <T> List<T> nullSafe(List<T> list) {
        List<T> emptyList = Collections.emptyList();
        
        return list == null ? emptyList : list;
    }
    
    /**
     * @return The empty Map if the input map is null, else the input map.
     */
    public static <S, T> Map<S, T> nullSafe(Map<S, T> map) {
        Map<S, T> emptyMap = Collections.emptyMap();
        
        return map == null ? emptyMap : map;
    }
    
    /**
     * Merge two <b>sorted</b> lists into a new sorted list.
     * Running time is linear in the combined size of the lists.
     */
    public static <T> List<T> merge(List<? extends T> list1, List<? extends T> list2, Comparator<? super T> comparator) {
        List<T> mergedList = new LinkedList<T>();
        
        Iterator<? extends T> i = list1.iterator();
        Iterator<? extends T> j = list2.iterator();
        
        T nextI = i.hasNext() ? i.next() : null;
        T nextJ = j.hasNext() ? j.next() : null;
        
        while (nextI != null || nextJ != null) {
            int comparison;
            
            if (nextI == null) {
                comparison = 1;
            }
            else if (nextJ == null) {
                comparison = -1;   
            }
            else {
                comparison = comparator.compare(nextI, nextJ);
            }
            
            if (comparison < 0) {
                mergedList.add(nextI);
                nextI = i.hasNext() ? i.next() : null;
            }
            else if (comparison > 0) {
                mergedList.add(nextJ);
                nextJ = j.hasNext() ? j.next() : null;
            }
            else {
                // nextI and nextJ are equal.  Pick nextI
                mergedList.add(nextI);
                nextI = i.hasNext() ? i.next() : null;
            }
        }
        
        return mergedList;
    }
    
    /**
     * Is the {@link Collection} sorted according to the given {@link Comparator}?
     */
    public static <T> boolean isSorted(Collection<T> collection, Comparator<? super T> comparator) {
        T previous = null;
        
        for (Iterator<T> i = collection.iterator(); i.hasNext();) {
            T current = i.next();
            if (previous != null && comparator.compare(previous, current) > 0) {
                return false;
            }
        }
        
        return true;
    }
    
}
