/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/util/Transformers.java#1 $
 *
 * Copyright (c) 2009 NetApp, Inc.
 * All rights reserved. 
 */
package com.netapp.common.util;

import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;

/**
 * Utility class to generate standard transformers.
 * @author sunday
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public class Transformers {
    private static final Transformer<Object, String> TO_STRING_TRANSFORMER = new ToStringTransformer();
    
    private Transformers() {
    }
    
    /**
     * Create a new transformer to reflectively read a class field in the tranform() method.
     * @param objClass Class of the objects this transformer works on
     * @param fieldClass Expected class of the values in field
     * @param field Field to be retrieved
     * @return New field transformer
     */
    public static <S, T> Transformer<S, T> fieldTransformer(Class<S> objClass, Class<T> fieldClass, Field field) {
        return new FieldTransformer<S, T>(objClass, fieldClass, field);
    }
    
    /**
     * Create a new transformer to reflectively read a class field in the tranform() method.
     * @param objClass Class of the objects this transformer works on
     * @param fieldClass Expected class of the values in field
     * @param fieldName Field to be retrieved
     * @return New field transformer
     * @throws UnsupportedOperationException The requested field could not be found.
     */
    public static <S, T> Transformer<S, T> fieldTransformer(Class<S> objClass, Class<T> fieldClass, String fieldName) {
        Field f = findField(objClass, fieldClass, fieldName);
        if (f != null) {
            return fieldTransformer(objClass, fieldClass, f);
        }
        throw new UnsupportedOperationException(fieldName);
    }
    
    /**
     * Create a new transformer to reflectively execute a method and return the result in the tranform() method.
     * @param objClass Class of the objects this transformer works on
     * @param returnClass Class of the return values from the supplied method
     * @param method Method on objClass to execute and return result
     * @param args Arguments to method for each invocation
     * @return New method transformer
     */
    public static <S, T> Transformer<S, T> methodTransformer(Class<S> objClass, Class<T> returnClass, Method method, Object... args) {
        return new MethodTransformer<S, T>(objClass, returnClass, method, args);
    }
    
    /**
     * Create a new transformer to reflectively execute a method and return the result in the tranform() method.
     * @param objClass Class of the objects this transformer works on
     * @param returnClass Class of the return values from the supplied method
     * @param methodName Method on objClass to execute and return result
     * @return New method transformer
     */
    public static <S, T> Transformer<S, T> methodTransformer(Class<S> objClass, Class<T> returnClass, String methodName) {
        Method method = findMethod(objClass, returnClass, methodName);
        if (method != null) {
            return methodTransformer(objClass, returnClass, method);
        }  
        throw new UnsupportedOperationException(methodName);
    }
    
    /**
     * Create a new transformer to reflectively retrieve a bean property of a class and return the result in the transform() method.
     * This will first attempt to find a valid Java Bean getter for the supplied property, if that fails it will access the field
     * directly.
     * @param objClass Class of the objects the new transformer works on
     * @param propertyClass Class of the requested property
     * @param propertyName Name of the requested property
     * @return New property transformer
     * @see #fieldTransformer(Class, Class, String)
     * @see #methodTransformer(Class, Class, String)
     */
    public static <S, T> Transformer<S, T> propertyTransformer(Class<S> objClass, Class<T> propertyClass, String propertyName) {
        String methodSuffix = StringUtils.initialUpperCase(propertyName);
        Method getter = findMethod(objClass, propertyClass, "get" + methodSuffix);
        if (getter == null && propertyClass.equals(Boolean.class)) {
            getter = findMethod(objClass, propertyClass, "is" + methodSuffix);
        }
        if (getter != null) {
            return methodTransformer(objClass, propertyClass, getter);
        }
        
        Field field = findField(objClass, propertyClass, propertyName);
        if (field != null) {
            return fieldTransformer(objClass, propertyClass, field);
        }
        throw new UnsupportedOperationException(propertyName);
    }
    
    /**
     * Create a new transformer which will transform an object to the value of its toString() method
     * @return New toString transformer
     */
    public static Transformer<Object, String> toStringTransformer() {
        return TO_STRING_TRANSFORMER;
    }
    
    /**
     * Transformer to reflectively invoke a method in transform()
     */
    private static class MethodTransformer<S, T> implements Transformer<S, T> {
        private final Method method;
        private final Class<T> returnClass;
        private final Object[] args;
        
        public MethodTransformer(Class<S> objClass, Class<T> returnClass, Method method, Object... args) {
            assert ClassUtils.typeAssignableFrom(returnClass, method.getReturnType());
            assert method.getParameterTypes().length == args.length;
            this.method = method;
            this.returnClass = returnClass;
            this.args = args;
            
            this.method.setAccessible(true);
        }

        public T transform(S arg0) {
            try {
                return returnClass.cast(method.invoke(arg0, args));
            }
            catch (IllegalArgumentException e) {
                throw new RuntimeException(e);
            }
            catch (IllegalAccessException e) {
                throw new RuntimeException(e);
            }
            catch (InvocationTargetException e) {
                throw new RuntimeException(e);
            }
        }
    }
    
    /**
     * Transformer to reflectively retrieve a field value
     */
    private static class FieldTransformer<S, T> implements Transformer<S, T> {
        private final Field field;
        private final Class<T> returnClass;
        
        public FieldTransformer(Class<S> objClass, Class<T> returnClass, Field field) {
            assert ClassUtils.typeAssignableFrom(returnClass, field.getType());
            this.field = field;
            this.returnClass = returnClass;
            this.field.setAccessible(true);
        }

        public T transform(S arg0) {
            try {
                return returnClass.cast(field.get(arg0));
            }
            catch (IllegalArgumentException e) {
                throw new RuntimeException(e);
            }
            catch (IllegalAccessException e) {
                throw new RuntimeException(e);
            }
        }
    }
    
    /**
     * Transformer to invoke toString()
     * @author sunday
     * @version $Revision: #1 $ $Date: 2010/08/20 $
     */
    private static class ToStringTransformer implements Transformer<Object, String> {
        public String transform(Object arg0) {
            return arg0.toString();
        }
    }
    
    /**
     * Attempt to find a 0 argument method in the given class, returning the given type.
     * @param clazz
     * @param type
     * @param name
     * @return Method of name returning type or null if no suitable method was found
     */
    private static Method findMethod(Class<?> clazz, Class<?> type, String name) {
        while (clazz != null) {
            for (Method m : clazz.getDeclaredMethods()) {
                if (m.getName().equals(name)
                        && !Modifier.isStatic(m.getModifiers())
                        && m.getParameterTypes().length == 0
                        && ClassUtils.typeAssignableFrom(type, m.getReturnType())) {
                    return m;
                }           
            }
            clazz = clazz.getSuperclass();
        }
        return null;
    }
    
    /**
     * Attempt to find a field in the given class of the given type
     * @param clazz
     * @param type
     * @param name
     * @return Field of name and type or null if no suitable field was found
     */
    private static Field findField(Class<?> clazz, Class<?> type, String name) {
        while (clazz != null) {
            for (Field f : clazz.getDeclaredFields()) {
                if (f.getName().equals(name)
                        && !Modifier.isStatic(f.getModifiers())
                        && ClassUtils.typeAssignableFrom(type, f.getType())) {
                    return f;
                }
            }
            clazz = clazz.getSuperclass();
        }
        return null;
    }
}

