/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/zapi/parser/Field.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.zapi.parser;

import java.math.BigInteger;
import java.util.Queue;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.netapp.common.MsgKey;
import com.netapp.common.logging.NALogger;
import com.netapp.common.zapi.generator.GeneratorFactory;

/**
 * Represents a single addressable field in a zapi api call or typedef.
 * 
 * @author sunday
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public class Field implements Comparable<Field> {
    private static final NALogger logger = NALogger.getLogger(Field.class);
    
    /**
     * Pattern to get a Range: [] from a description. This only does a first pass. 
     * @see #PARSE_RANGE_PATTERN
     */
    static final Pattern DESC_RANGE_PATTERN = 
        Pattern.compile("Range\\s*:\\s*(\\[[^\\]]+\\])");
    
    /**
     * Matches a Range argument min[ws]separator[ws]max
     * min ::= range_arg
     * max ::= range_arg
     * range_arg ::= literal | power2
     * literal ::= [negate]{digit}
     * power2 :: = [negate]["2^"{digit}["-1"]
     * separator ::= "-" | ".."
     */
    static final Pattern PARSE_RANGE_PATTERN = 
        Pattern.compile("(?:^|\\[)(\\-?(?:\\d+|2\\^\\d+(?:-1)?))\\s*(?:-|\\.\\.)\\s*(\\-?(?:\\d+|2\\^\\d+(?:-1)?))(?:\\]|$)");
    
    /**
     * Tear apart the individual range_arg construct
     * 1: literal (null if exponent)
     * 2: "-" (null if literal or not negative)
     * 3: exponent (null if literal)
     * 4: "-1" or null
     * 
     */
    static final Pattern PARSE_RANGE_ARG = 
        Pattern.compile("(-?\\d+)|(-)?2\\^(\\d+)(-1)?");
    static final int PARSE_RANGE_LITERAL_GROUP = 1;
    static final int PARSE_RANGE_EXP_NEGATE_GROUP = 2;
    static final int PARSE_RANGE_EXP_GROUP = 3;
    static final int PARSE_RANGE_EXP_MIN_ONE_GROUP = 4;
    
    /**
     * Check for a sign modifier on a type.
     * This is to work around a bug in the Ontap docs
     */
    static final Pattern SIGN_MODIFIER = Pattern.compile("((?:un)signed)\\s+(\\S+)");
    
    /** The declared zapi name of this field */
    String name;
    
    /** The description of this field */
    String desc;
    
    /** The declared zapi type of this field */
    String type;
    
    /** Is this declared as a collection */
    boolean collection; 
    
    /** The allowed range of this field for integers */
    Range range;
    
    /** If this field has an optional declaration on it */
    boolean optional;
    
    /** The Api or Typedef this field is declared in */
    Type declaringType;
    
    /**
     * Construct a new Field
     * @param declaringType The Api or Typedef defining this field
     * @param name The declared zapi name of this field
     */
    public Field(Type declaringType, String name) {
        this.declaringType = declaringType;
        this.name = name;
    }
    
    /**
     * Get the declared zapi name of this field
     * @return declared zapi name of this field
     */
    public String getName() {
        return this.name;
    }
    
    /**
     * Set the description of this field. Additionally, if the description contains what looks
     * like a legal range declaration, set the range as well.
     * @param desc declared description of this field
     */
    protected void setDesc(String desc) {
        this.desc = desc;
        Matcher m = DESC_RANGE_PATTERN.matcher(desc);
        if (m.find()) {
            this.setRange(m.group(1));
        }
    }
    
    /**
     * Get the description of this field
     * @return description of this field
     */
    public String getDesc() {
        return this.desc;
    }
    
    /**
     * Get the range of this field
     * @return Allowed range of this field, or null if none has been declared
     */
    public Range getRange() {
        return this.range;
    }
    
    /**
     * Set the allowed range of this field.
     * @param range Allowed range of this field, in [X..Y] format
     * @see #PARSE_RANGE_PATTERN for allowed range formats
     */
    private void setRange(String range) {
        range = range.replaceAll("[\\(\\)\\s]", "");
        Matcher m = PARSE_RANGE_PATTERN.matcher(range);

        if (m.find()) {
            this.range = new Range(m.group(1), m.group(2));
        }
        else {
            declaringType.parser.appendWarning(MsgKey.ZAPI_GENERATOR_INVALID_RANGE, range, this.name);
        }
    }
    
    /**
     * Set the zapi type of this field
     * @param type zapi type of this field. Must be one of the simple types "boolean", 
     * "integer", "string", or a zapi typedef in the same namespace.
     */
    private void setType(String type) {
        type = type.trim().toLowerCase();
        
        Matcher signMatcher = SIGN_MODIFIER.matcher(type);
        if (signMatcher.matches()) {
            declaringType.parser.appendWarning(MsgKey.ZAPI_GENERATOR_FIELD_MODIFIER_INVALID, signMatcher.group(1), this.name);
            type = signMatcher.group(2);
        }
        
        boolean newCollection = false;
        if (type.endsWith("[]")) {
            newCollection = true;
            type = type.substring(0, type.length() - 2);
        }
        
        if (this.type == null) {
            this.type = type;
            this.collection = newCollection;
        }
        else if (this.type.equals(type) && this.collection == newCollection) {
            // Silently ignore
        }
        else {
            declaringType.parser.appendWarning(MsgKey.ZAPI_GENERATOR_FIELD_TYPE_REDEFINED, declaringType, this.name, this.type, type);
        }
    }
    
    /** 
     * Return the Api or Typedef this field is declared in 
     */
    public Type getDeclaringType() {
        return this.declaringType;
    }
    
    /**
     * Return the zapi type of this field
     * @return zapi type of this field
     */
    public String getType() {
        return this.type;
    }
    
    public Type getTypedefType() {
        // See if we can find it in the current namespace
        return this.declaringType.getNamespace().getType(this.type);   
    }
    
    /**
     * Check if this field represents a collection
     * @return true if this field is a collection of objects
     */
    public boolean isCollection() {
        return collection;
    }
    
    /**
     * Populate this field with tags from the supplied queue,
     * stopping on the first unknown tag. Any processed tags are
     * removed from the queue.
     * @param q Queue of tags to parse
     */
    void populate(Queue<Tag> q) {
        logger.internal("Processing Field " + name);
        Tag tag = q.peek();
        while (tag != null) {
            if (tag.is(TagEnum.DESC)) {
                this.setDesc(tag.getValue());
            }
            else if (tag.is(TagEnum.TYPE)) {
                String[] words = tag.getValue().split(",");
                this.setType(words[0]);
                this.type = this.type.trim();
                for (int i = 1; i < words.length; i++) {
                    if (words[i].trim().equalsIgnoreCase("optional")) {
                        this.optional = true;
                    }
                }
            }
            else if (tag.is(TagEnum.RANGE)) {
                this.setRange(tag.getValue());
            }
            else if (tag.is(TagEnum.ATTRIBUTE)) {
                // Not handling this
            }
            else {
                // Tag not recognized
                logger.internal("Stopping Field on " + tag.getType());
                break;
            }
            // Remove the one we were looking at
            q.poll();
            
            // Peek at the next one
            tag = q.peek();
        }
    }

    /**
     * Check if this field is optional or required
     * @return true if this field is optional per the zapi specs
     */
    public boolean isOptional() {
        return this.optional;
    }
    
    /**
     * Check if we should generate a an EnclosingType withXXX setter for 
     * this field
     * @return true if this field should have an additional withXXX setter
     */
    public boolean isChainable() {
        return true;
    }
    
    public void addWordReplacements(GeneratorFactory<?, ?, ?, ?> generatorFactory) {
        generatorFactory.getDescriptionGenerator().addWordReplacement(
            this.name,
            generatorFactory.getFieldGenerator(this).getNameGenerated());
    }
    
    
    /**
     * Compare this Field to another based on the java field name
     * @see java.lang.Comparable#compareTo(java.lang.Object)
     */
    public int compareTo(Field obj) {
        return this.name.compareTo(obj.name);
    }
    
    /**
     * Print a string representation of this field
     * @return representation of this field
     * @see java.lang.Object#toString()
     */
    @Override
    public String toString() {
        return this.type + " " + this.name;
    }
    
    public class Range {
        /** Minumum value for this field */
        private final String min;
        
        /** Maximum value for this field */
        private final String max;
        
        private final BigInteger minInteger;
        private final BigInteger maxInteger;
        
        /**
         * Construct a new range from the supplied minium and maximum values.
         * @param min Minimum value for this field
         * @param max Maximum value for this field
         * @param field TODO
         */
        Range(String min, String max) {
            this.min = min;
            this.minInteger = expand(min);
            this.max = max;
            this.maxInteger = expand(max);
        }
        
        public String getMin() {
            return this.min;
        }
        
        public String getMax() {
            return this.max;
        }
        
        public BigInteger getMinInteger() {
            return this.minInteger;
        }
        
        public BigInteger getMaxInteger() {
            return this.maxInteger;
        }
        
        public int getBitCount() {
            // bitLength() doesn't include the sign bit
            return Math.max(minInteger.bitLength(), maxInteger.bitLength()) + 1;
        }
        
        /**
         * Expand a valid ZAPI range representation into a BigIntege.
         * @param num Number to calculate, either as a literal number or a power-of-two.
         * See {@link Field#PARSE_RANGE_PATTERN} for legal representations.
         * @return
         */
        private BigInteger expand(String num) {
            Matcher m = Field.PARSE_RANGE_ARG.matcher(num);
            BigInteger result;
            
            if (m.matches()) {
                if (m.group(1) != null) {
                    // Literal number
                    result = new BigInteger(m.group(1));
                }
                else {
                    int exponent = Integer.parseInt(m.group(Field.PARSE_RANGE_EXP_GROUP));
                    result = new BigInteger("2").pow(exponent);
                    if (m.group(Field.PARSE_RANGE_EXP_NEGATE_GROUP) != null) {
                        result = result.negate();
                    }
                    if (m.group(Field.PARSE_RANGE_EXP_MIN_ONE_GROUP) != null) {
                        result = result.subtract(BigInteger.ONE);
                    }
                }
            }
            else {
                // TODO: Throw an assertion here
                // This should not happen, errors should have been caught earlier
                getDeclaringType().parser.appendError(MsgKey.ZAPI_GENERATOR_INVALID_RANGE_COMPONENT, num, Field.this.name);
                result = BigInteger.ZERO;
            }
            return result;
        }
    }
}
