/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/zapi/parser/Typedef.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.zapi.parser;

import java.util.Queue;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.netapp.common.MsgKey;
import com.netapp.common.logging.NALogger;

/**
 * Represents zapidoc declared typedef struct of elements or aliase
 * @author sunday
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public class Typedef extends Type {
    private static final NALogger logger = NALogger.getLogger(Typedef.class);
    
    /** Pattern matching zapi-type, modifiers on a type tag */
    private static final Pattern TYPE_MODIFIER_PATTERN = Pattern.compile("(\\S+)\\s*,\\s*(.*)");
    
    /** 
     * Type of the typedef if it's just an alias.
     * A typedef may have elements or a type, but not both.
     */
    private String type;
    
    /**
     * Construct a new typedef
     * @param category category of the new typedef
     * @param name zapi name of this type
     */
    Typedef(ZapidocParser parser, String name) {
        super(parser, name);
    }
    
    public String getType() {
        return this.type;
    }
    
    /**
     * Retrieve the field with the given zapi name
     * @param fieldName zapi name of the field to find
     * @return Field with the specified zapi name or null if not found
     */
    protected TypedefField getField(String fieldName) {
        TypedefField field = null;
        for (Field f : fields) {
            if (fieldName.equals(f.getName())) {
                // The collection should be cast safe for us
                field = (TypedefField)f;
                break;
            }
        }
        return field;      
    }
    
    /**
     * 
     * @see com.netapp.common.zapi.parser.Type#populate(java.util.Queue)
     */
    @Override
    public void populate(Queue<Tag> q) {
        logger.internal("Processing Field " + name);
        Tag tag = q.peek();
        while (tag != null) {
            TypedefField field = null;
            if (tag.is(TagEnum.DESC)) {
                this.setDesc(tag.getValue());
            }
            else if (tag.is(TagEnum.TYPE)) {
                if (fields.isEmpty()) {
                    String destinationType = tag.getValue().toLowerCase();
                    
                    // Catch burt 246082
                    Matcher matcher = TYPE_MODIFIER_PATTERN.matcher(destinationType);
                    if (matcher.matches()) {
                        this.parser.appendWarning(MsgKey.ZAPI_GENERATOR_TYPEDEF_ALIAS_MODIFER, this, matcher.group(2));
                        destinationType = matcher.group(1);
                    }
                    
                    if (SimpleType.SIMPLE_TYPES.containsKey(destinationType)) {
                        this.type = destinationType;
                    }
                    else {
                        this.parser.appendWarning(MsgKey.ZAPI_GENERATOR_INVALID_TYPEDEF_ALIAS_TYPE, destinationType, this);
                        this.type = destinationType;
                    }
                }
                else {
                    this.parser.appendError(MsgKey.ZAPI_GENERATOR_INVALID_TYPEDEF_ALIAS, this);
                }
            }
            else if (tag.is(TagEnum.ELEMENT)) {
                if (type == null) {
                    field = getField(tag.getFirstWord());
                    if (field == null) {
                        field = new TypedefField(tag.getFirstWord());
                        this.fields.add(field);
                    }
                    else {
                        this.parser.appendError(MsgKey.ZAPI_GENERATOR_API_FIELD_DEFINED_TWICE, this, field.name);
                    } 
                }
                else { 
                    this.parser.appendError(MsgKey.ZAPI_GENERATOR_INVALID_TYPEDEF_ALIAS, this);
                }
            }
            else if (tag.is(TagEnum.NAMESPACE)) {
                this.setNamespace(tag.getFirstWord());
            }
            else if (tag.is(TagEnum.ATTRIBUTE)) {
                // Not handling this
                logger.info(MsgKey.ZAPI_GENERATOR_UNHANDLED_TYPEDEF_ATTRIBUTE, tag.getFirstWord(), this);
            }
            else {
                // Tag not recognized
                logger.internal("Stopping Field on " + tag.getType());
                break;
            }
            // Remove the one we were looking at
            q.poll();
            
            // This is kind of hacky, but we want to process this after we
            // consume the tag that created it.
            if (field != null) {
                field.populate(q);
            }
            // Peek at the next one
            tag = q.peek();
        }
        if (this.namespace == null) {
            // Typedefs should always be in the current namespace
            this.setNamespace(parser.getLastNamespace());
        }
    }

    /**
     * 
     * @see com.netapp.common.zapi.parser.Type#isAlias()
     */
    @Override
    public boolean isAlias() {
        return this.type != null;
    }

    /**
     * If this typedef is an alias, then return the root
     * of the alias hierarchy.
     * 
     * If this typedef is not an alias, then return this.
     */
    public Type getRootType() {
        if (this.type == null) {
            return this;
        }
        
        return this.namespace.getType(this.type).getRootType();
    }

    /**
     * Field extended to represent a zapidoc element
     * @author sunday
     * @version $Revision: #1 $ $Date: 2010/08/20 $
     */
    public class TypedefField extends Field {
        /**
         * Construct a new field in the current Typedef
         * @param name zapi name of new field
         */
        public TypedefField(String name) {
            super(Typedef.this, name);
        }
        
    }
}